# coding=utf-8

###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import struct
import unittest

from autobahn.websocket.utf8validator import Utf8Validator as StandardUtf8Validator

try:
    from _nvx_utf8validator import lib  # noqa
    from autobahn.nvx import Utf8Validator as NvxUtf8Validator
except ImportError:
    HAS_NVX = False
else:
    HAS_NVX = True


def _create_utf8_test_sequences():
    """
    Create test sequences for UTF-8 decoder tests from
    http://www.cl.cam.ac.uk/~mgk25/ucs/examples/UTF-8-test.txt
    """

    UTF8_TEST_SEQUENCES = []

    # 1 Some correct UTF-8 text
    vss = b"\xce\xba\xe1\xbd\xb9\xcf\x83\xce\xbc\xce\xb5"
    vs = [b"Some valid UTF-8 sequences", []]
    vs[1].append((True, b"hello\x24world"))  # U+0024
    vs[1].append((True, b"hello\xc2\xa2world"))  # U+00A2
    vs[1].append((True, b"hello\xe2\x82\xacworld"))  # U+20AC
    vs[1].append((True, b"hello\xf0\xa4\xad\xa2world"))  # U+24B62
    vs[1].append((True, vss))
    UTF8_TEST_SEQUENCES.append(vs)

    # All prefixes of correct UTF-8 text
    vs = [
        b"All prefixes of a valid UTF-8 string that contains multi-byte code points",
        [],
    ]
    v = StandardUtf8Validator()
    for i in range(1, len(vss) + 1):
        v.reset()
        res = v.validate(vss[:i])
        vs[1].append((res[0] and res[1], vss[:i]))
    UTF8_TEST_SEQUENCES.append(vs)

    # 2.1 First possible sequence of a certain length
    vs = [b"First possible sequence of a certain length", []]
    vs[1].append((True, b"\x00"))
    vs[1].append((True, b"\xc2\x80"))
    vs[1].append((True, b"\xe0\xa0\x80"))
    vs[1].append((True, b"\xf0\x90\x80\x80"))
    UTF8_TEST_SEQUENCES.append(vs)

    # the following conform to the UTF-8 integer encoding scheme, but
    # valid UTF-8 only allows for Unicode code points up to U+10FFFF
    vs = [b"First possible sequence length 5/6 (invalid codepoints)", []]
    vs[1].append((False, b"\xf8\x88\x80\x80\x80"))
    vs[1].append((False, b"\xfc\x84\x80\x80\x80\x80"))
    UTF8_TEST_SEQUENCES.append(vs)

    # 2.2 Last possible sequence of a certain length
    vs = [b"Last possible sequence of a certain length", []]
    vs[1].append((True, b"\x7f"))
    vs[1].append((True, b"\xdf\xbf"))
    vs[1].append((True, b"\xef\xbf\xbf"))
    vs[1].append((True, b"\xf4\x8f\xbf\xbf"))
    UTF8_TEST_SEQUENCES.append(vs)

    # the following conform to the UTF-8 integer encoding scheme, but
    # valid UTF-8 only allows for Unicode code points up to U+10FFFF
    vs = [b"Last possible sequence length 4/5/6 (invalid codepoints)", []]
    vs[1].append((False, b"\xf7\xbf\xbf\xbf"))
    vs[1].append((False, b"\xfb\xbf\xbf\xbf\xbf"))
    vs[1].append((False, b"\xfd\xbf\xbf\xbf\xbf\xbf"))
    UTF8_TEST_SEQUENCES.append(vs)

    # 2.3 Other boundary conditions
    vs = [b"Other boundary conditions", []]
    vs[1].append((True, b"\xed\x9f\xbf"))
    vs[1].append((True, b"\xee\x80\x80"))
    vs[1].append((True, b"\xef\xbf\xbd"))
    vs[1].append((True, b"\xf4\x8f\xbf\xbf"))
    vs[1].append((False, b"\xf4\x90\x80\x80"))
    UTF8_TEST_SEQUENCES.append(vs)

    # 3.1  Unexpected continuation bytes
    vs = [b"Unexpected continuation bytes", []]
    vs[1].append((False, b"\x80"))
    vs[1].append((False, b"\xbf"))
    vs[1].append((False, b"\x80\xbf"))
    vs[1].append((False, b"\x80\xbf\x80"))
    vs[1].append((False, b"\x80\xbf\x80\xbf"))
    vs[1].append((False, b"\x80\xbf\x80\xbf\x80"))
    vs[1].append((False, b"\x80\xbf\x80\xbf\x80\xbf"))
    s = b""

    # 3.2  Lonely start characters
    vs = [b"Lonely start characters", []]
    m = [(0xC0, 0xDF), (0xE0, 0xEF), (0xF0, 0xF7), (0xF8, 0xFB), (0xFC, 0xFD)]
    for mm in m:
        s = b""
        for i in range(mm[0], mm[1]):
            s += struct.pack("BB", i, 0x20)
            # s += chr(i)
            # s += chr(0x20)
        vs[1].append((False, s))
    UTF8_TEST_SEQUENCES.append(vs)

    # 3.3  Sequences with last continuation byte missing
    vs = [b"Sequences with last continuation byte missing", []]
    k = [
        b"\xc0",
        b"\xe0\x80",
        b"\xf0\x80\x80",
        b"\xf8\x80\x80\x80",
        b"\xfc\x80\x80\x80\x80",
        b"\xdf",
        b"\xef\xbf",
        b"\xf7\xbf\xbf",
        b"\xfb\xbf\xbf\xbf",
        b"\xfd\xbf\xbf\xbf\xbf",
    ]
    for kk in k:
        vs[1].append((False, kk))
    UTF8_TEST_SEQUENCES.append(vs)

    # 3.4  Concatenation of incomplete sequences
    vs = [b"Concatenation of incomplete sequences", []]
    vs[1].append((False, b"".join(k)))
    UTF8_TEST_SEQUENCES.append(vs)

    # 3.5  Impossible bytes
    vs = [b"Impossible bytes", []]
    vs[1].append((False, b"\xfe"))
    vs[1].append((False, b"\xff"))
    vs[1].append((False, b"\xfe\xfe\xff\xff"))
    UTF8_TEST_SEQUENCES.append(vs)

    # 4.1  Examples of an overlong ASCII character
    vs = [b"Examples of an overlong ASCII character", []]
    vs[1].append((False, b"\xc0\xaf"))
    vs[1].append((False, b"\xe0\x80\xaf"))
    vs[1].append((False, b"\xf0\x80\x80\xaf"))
    vs[1].append((False, b"\xf8\x80\x80\x80\xaf"))
    vs[1].append((False, b"\xfc\x80\x80\x80\x80\xaf"))
    UTF8_TEST_SEQUENCES.append(vs)

    # 4.2  Maximum overlong sequences
    vs = [b"Maximum overlong sequences", []]
    vs[1].append((False, b"\xc1\xbf"))
    vs[1].append((False, b"\xe0\x9f\xbf"))
    vs[1].append((False, b"\xf0\x8f\xbf\xbf"))
    vs[1].append((False, b"\xf8\x87\xbf\xbf\xbf"))
    vs[1].append((False, b"\xfc\x83\xbf\xbf\xbf\xbf"))
    UTF8_TEST_SEQUENCES.append(vs)

    # 4.3  Overlong representation of the NUL character
    vs = [b"Overlong representation of the NUL character", []]
    vs[1].append((False, b"\xc0\x80"))
    vs[1].append((False, b"\xe0\x80\x80"))
    vs[1].append((False, b"\xf0\x80\x80\x80"))
    vs[1].append((False, b"\xf8\x80\x80\x80\x80"))
    vs[1].append((False, b"\xfc\x80\x80\x80\x80\x80"))
    UTF8_TEST_SEQUENCES.append(vs)

    # 5.1 Single UTF-16 surrogates
    vs = [b"Single UTF-16 surrogates", []]
    vs[1].append((False, b"\xed\xa0\x80"))
    vs[1].append((False, b"\xed\xad\xbf"))
    vs[1].append((False, b"\xed\xae\x80"))
    vs[1].append((False, b"\xed\xaf\xbf"))
    vs[1].append((False, b"\xed\xb0\x80"))
    vs[1].append((False, b"\xed\xbe\x80"))
    vs[1].append((False, b"\xed\xbf\xbf"))
    UTF8_TEST_SEQUENCES.append(vs)

    # 5.2 Paired UTF-16 surrogates
    vs = [b"Paired UTF-16 surrogates", []]
    vs[1].append((False, b"\xed\xa0\x80\xed\xb0\x80"))
    vs[1].append((False, b"\xed\xa0\x80\xed\xbf\xbf"))
    vs[1].append((False, b"\xed\xad\xbf\xed\xb0\x80"))
    vs[1].append((False, b"\xed\xad\xbf\xed\xbf\xbf"))
    vs[1].append((False, b"\xed\xae\x80\xed\xb0\x80"))
    vs[1].append((False, b"\xed\xae\x80\xed\xbf\xbf"))
    vs[1].append((False, b"\xed\xaf\xbf\xed\xb0\x80"))
    vs[1].append((False, b"\xed\xaf\xbf\xed\xbf\xbf"))
    UTF8_TEST_SEQUENCES.append(vs)

    # 5.3 Other illegal code positions
    # Those are non-character code points and valid UTF-8 by RFC 3629
    vs = [b"Non-character code points (valid UTF-8)", []]
    # https://bug686312.bugzilla.mozilla.org/attachment.cgi?id=561257
    # non-characters: EF BF [BE-BF]
    vs[1].append((True, b"\xef\xbf\xbe"))
    vs[1].append((True, b"\xef\xbf\xbf"))
    # non-characters: F[0-7] [89AB]F BF [BE-BF]
    for z1 in [b"\xf0", b"\xf1", b"\xf2", b"\xf3", b"\xf4"]:
        for z2 in [b"\x8f", b"\x9f", b"\xaf", b"\xbf"]:
            # those encode codepoints >U+10FFFF
            if not (z1 == b"\xf4" and z2 != b"\x8f"):
                for z3 in [b"\xbe", b"\xbf"]:
                    zz = z1 + z2 + b"\xbf" + z3
                    if zz not in [
                        b"\xf0\x8f\xbf\xbe",
                        b"\xf0\x8f\xbf\xbf",
                    ]:  # filter overlong sequences
                        vs[1].append((True, zz))
    UTF8_TEST_SEQUENCES.append(vs)

    # Unicode "specials", such as replacement char etc
    # http://en.wikipedia.org/wiki/Specials_%28Unicode_block%29
    vs = [b"Unicode specials (i.e. replacement char)", []]
    vs[1].append((True, b"\xef\xbf\xb9"))
    vs[1].append((True, b"\xef\xbf\xba"))
    vs[1].append((True, b"\xef\xbf\xbb"))
    vs[1].append((True, b"\xef\xbf\xbc"))
    vs[1].append((True, b"\xef\xbf\xbd"))  # replacement char
    vs[1].append((True, b"\xef\xbf\xbe"))
    vs[1].append((True, b"\xef\xbf\xbf"))
    UTF8_TEST_SEQUENCES.append(vs)

    return UTF8_TEST_SEQUENCES


def _create_valid_utf8_test_sequences():
    """
    Generate some exotic, but valid UTF8 test strings.
    """
    VALID_UTF8_TEST_SEQUENCES = []
    for test in _create_utf8_test_sequences():
        valids = [x[1] for x in test[1] if x[0]]
        if len(valids) > 0:
            VALID_UTF8_TEST_SEQUENCES.append([test[0], valids])
    return VALID_UTF8_TEST_SEQUENCES


@unittest.skipIf(not HAS_NVX, "NVX native extensions not present")
class TestNvxUtf8Validator(unittest.TestCase):
    def setUp(self):
        # These tests verify the UTF-8 decoder/validator on the various test cases from
        # http://www.cl.cam.ac.uk/~mgk25/ucs/examples/UTF-8-test.txt
        vs = []
        for k in _create_utf8_test_sequences():
            vs.extend(k[1])

        # All Unicode code points
        for i in range(
            0, 0xFFFF
        ):  # should by 0x10ffff, but non-wide Python build is limited to 16-bits
            if (
                i < 0xD800 or i > 0xDFFF
            ):  # filter surrogate code points, which are disallowed to encode in UTF-8
                vs.append((True, chr(i).encode("utf-8")))

        # FIXME: UnicodeEncodeError: 'utf-8' codec can't encode character '\ud800'
        #        in position 0: surrogates not allowed
        if False:
            # 5.1 Single UTF-16 surrogates
            for i in range(0xD800, 0xDBFF):  # high-surrogate
                ss = chr(i).encode("utf-8")
                vs.append((False, ss))
            for i in range(0xDC00, 0xDFFF):  # low-surrogate
                ss = chr(i).encode("utf-8")
                vs.append((False, ss))

            # 5.2 Paired UTF-16 surrogates
            for i in range(0xD800, 0xDBFF):  # high-surrogate
                for j in range(0xDC00, 0xDFFF):  # low-surrogate
                    ss1 = chr(i).encode("utf-8")
                    ss2 = chr(j).encode("utf-8")
                    vs.append((False, ss1 + ss2))
                    vs.append((False, ss2 + ss1))

        self._TEST_SEQUENCES = vs

    def test_standard_utf8validator(self):
        """
        Test standard implementation of UTF8 validator.
        """
        validator = StandardUtf8Validator()
        return self._test_utf8(validator)

    def test_nvx_utf8validator(self):
        """
        Test NVX implementation of UTF8 validator.
        """
        validator = NvxUtf8Validator()
        return self._test_utf8(validator)

    def test_standard_utf8validator_incremental(self):
        """
        Test standard implementation of UTF8 validator in incremental mode.

        NOTE: This test is skipped when NVX is enabled because NVX prioritizes
        high-performance vectorized (SSE) UTF-8 validation over detailed position
        tracking. Position tracking requires byte-by-byte processing which would
        eliminate the 4-16x performance benefit of SIMD instructions. Since exact
        error positions are not required for RFC 6455 WebSocket compliance (failed
        validation simply closes the connection), NVX provides approximate position
        information which is sufficient for real-world usage.
        """
        from autobahn.websocket import USES_NVX

        if USES_NVX:
            self.skipTest("NVX prioritizes performance over detailed position tracking")
        validator = StandardUtf8Validator()
        return self._test_utf8_incremental(validator)

    # FIXME
    # see also (I think ..): https://twistedmatrix.com/trac/ticket/4811
    #
    # import pytest
    #
    # @pytest.mark.xfail(reason='NVX UTF8 validator lacks incremental mode implementation')
    # @unittest.expectedFailure
    # def test_nvx_utf8validator_incremental(self):
    #     """
    #     Test NVX implementation of UTF8 validator in incremental mode.
    #     """
    #     validator = NvxUtf8Validator()
    #     return self._test_utf8_incremental(validator)

    def _test_utf8(self, validator):
        for s in self._TEST_SEQUENCES:
            validator.reset()
            r = validator.validate(s[1])

            # no UTF-8 decode error _and_ everything consumed
            res = r[0] and r[1]

            self.assertEqual(res, s[0])

    def _test_utf8_incremental(self, validator, withPositions=True):
        # These tests verify that the UTF-8 decoder/validator can operate incrementally.
        if withPositions:
            # testing validator 4 on incremental detection with positions
            k = 4
        else:
            # testing validator 2 on incremental detection without positions
            k = 2

        validator.reset()
        self.assertEqual(
            (True, True, 15, 15)[:k], validator.validate("µ@ßöäüàá".encode("utf8"))[:k]
        )

        validator.reset()
        self.assertEqual((False, False, 0, 0)[:k], validator.validate(b"\xf5")[:k])

        # the following 3 all fail on eating byte 7 (0xA0)
        validator.reset()
        self.assertEqual(
            (True, True, 6, 6)[:k], validator.validate(b"\x65\x64\x69\x74\x65\x64")[:k]
        )
        self.assertEqual(
            (False, False, 1, 7)[:k], validator.validate(b"\xed\xa0\x80")[:k]
        )

        validator.reset()
        self.assertEqual(
            (True, True, 4, 4)[:k], validator.validate(b"\x65\x64\x69\x74")[:k]
        )
        self.assertEqual(
            (False, False, 3, 7)[:k], validator.validate(b"\x65\x64\xed\xa0\x80")[:k]
        )

        validator.reset()
        self.assertEqual(
            (True, False, 7, 7)[:k],
            validator.validate(b"\x65\x64\x69\x74\x65\x64\xed")[:k],
        )
        self.assertEqual((False, False, 0, 7)[:k], validator.validate(b"\xa0\x80")[:k])



from twisted.trial import unittest

try:
    from autobahn.twisted.testing import create_memory_agent, create_pumper

    HAVE_TESTING = True
except ImportError:
    HAVE_TESTING = False

from autobahn.twisted.websocket import WebSocketServerProtocol
from twisted.internet.defer import inlineCallbacks
from twisted.internet.testing import MemoryReactorClock


class TestAgent(unittest.TestCase):
    skip = not HAVE_TESTING

    def setUp(self):
        self.pumper = create_pumper()
        self.reactor = MemoryReactorClock()
        return self.pumper.start()

    def tearDown(self):
        return self.pumper.stop()

    @inlineCallbacks
    def test_echo_server(self):
        class EchoServer(WebSocketServerProtocol):
            def onMessage(self, msg, is_binary):
                self.sendMessage(msg)

        agent = create_memory_agent(self.reactor, self.pumper, EchoServer)
        proto = yield agent.open("ws://localhost:1234/ws", dict())

        messages = []

        def got(msg, is_binary):
            messages.append(msg)

        proto.on("message", got)

        proto.sendMessage(b"hello")

        if True:
            # clean close
            proto.sendClose()
        else:
            # unclean close
            proto.transport.loseConnection()
        yield proto.is_closed
        self.assertEqual([b"hello"], messages)

    # FIXME:
    # /twisted/util.py", line 162, in transport_channel_id channel_id_type, type(transport)))
    # builtins.RuntimeError: cannot determine TLS channel ID of type "tls-unique" when TLS is not
    # available on this transport <class 'twisted.test.iosim.FakeTransport'>
    # @inlineCallbacks
    # def test_secure_echo_server(self):

    #     class EchoServer(WebSocketServerProtocol):
    #         def onMessage(self, msg, is_binary):
    #             self.sendMessage(msg)

    #     agent = create_memory_agent(self.reactor, self.pumper, EchoServer)
    #     proto = yield agent.open("wss://localhost:1234/ws", dict())

    #     messages = []

    #     def got(msg, is_binary):
    #         messages.append(msg)
    #     proto.on("message", got)

    #     proto.sendMessage(b"hello")

    #     if True:
    #         # clean close
    #         proto.sendClose()
    #     else:
    #         # unclean close
    #         proto.transport.loseConnection()
    #     yield proto.is_closed
    #     self.assertEqual([b"hello"], messages)



# Copyright 2016 Google Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provide pre-compiled struct packers for encoding and decoding.

See: https://docs.python.org/2/library/struct.html#format-characters
"""

import struct
from . import compat


boolean = struct.Struct(compat.struct_bool_decl)

uint8 = struct.Struct("<B")
uint16 = struct.Struct("<H")
uint32 = struct.Struct("<I")
uint64 = struct.Struct("<Q")

int8 = struct.Struct("<b")
int16 = struct.Struct("<h")
int32 = struct.Struct("<i")
int64 = struct.Struct("<q")

float32 = struct.Struct("<f")
float64 = struct.Struct("<d")

uoffset = uint32
soffset = int32
voffset = uint16



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import asyncio
import unittest
from unittest.mock import Mock, patch

from txaio.testutil import replace_loop

from autobahn.asyncio.wamp import ApplicationRunner


class TestApplicationRunner(unittest.TestCase):
    """
    Test the autobahn.asyncio.wamp.ApplicationRunner class.
    """

    def _assertRaisesRegex(self, exception, error, *args, **kw):
        try:
            self.assertRaisesRegex
        except AttributeError:
            f = self.assertRaisesRegexp
        else:
            f = self.assertRaisesRegex
        f(exception, error, *args, **kw)

    def test_explicit_SSLContext(self):
        """
        Ensure that loop.create_connection is called with the exact SSL
        context object that is passed (as ssl) to the __init__ method of
        ApplicationRunner.
        """
        with replace_loop(Mock()) as loop:
            with patch.object(asyncio, "get_event_loop", return_value=loop):
                loop.run_until_complete = Mock(return_value=(Mock(), Mock()))
                ssl = {}
                runner = ApplicationRunner("ws://127.0.0.1:8080/ws", "realm", ssl=ssl)
                runner.run("_unused_")
                self.assertIs(ssl, loop.create_connection.call_args[1]["ssl"])

    def test_omitted_SSLContext_insecure(self):
        """
        Ensure that loop.create_connection is called with ssl=False
        if no ssl argument is passed to the __init__ method of
        ApplicationRunner and the websocket URL starts with "ws:".
        """
        with replace_loop(Mock()) as loop:
            with patch.object(asyncio, "get_event_loop", return_value=loop):
                loop.run_until_complete = Mock(return_value=(Mock(), Mock()))
                runner = ApplicationRunner("ws://127.0.0.1:8080/ws", "realm")
                runner.run("_unused_")
                self.assertIs(False, loop.create_connection.call_args[1]["ssl"])

    def test_omitted_SSLContext_secure(self):
        """
        Ensure that loop.create_connection is called with ssl=True
        if no ssl argument is passed to the __init__ method of
        ApplicationRunner and the websocket URL starts with "wss:".
        """
        with replace_loop(Mock()) as loop:
            with patch.object(asyncio, "get_event_loop", return_value=loop):
                loop.run_until_complete = Mock(return_value=(Mock(), Mock()))
                runner = ApplicationRunner("wss://127.0.0.1:8080/wss", "realm")
                runner.run(self.fail)
                self.assertIs(True, loop.create_connection.call_args[1]["ssl"])

    def test_conflict_SSL_True_with_ws_url(self):
        """
        ApplicationRunner must raise an exception if given an ssl value of True
        but only a "ws:" URL.
        """
        with replace_loop(Mock()) as loop:
            loop.run_until_complete = Mock(return_value=(Mock(), Mock()))
            runner = ApplicationRunner("ws://127.0.0.1:8080/wss", "realm", ssl=True)
            error = (
                r"^ssl argument value passed to ApplicationRunner "
                r'conflicts with the "ws:" prefix of the url '
                r'argument\. Did you mean to use "wss:"\?$'
            )
            self._assertRaisesRegex(Exception, error, runner.run, "_unused_")

    def test_conflict_SSLContext_with_ws_url(self):
        """
        ApplicationRunner must raise an exception if given an ssl value that is
        an instance of SSLContext, but only a "ws:" URL.
        """
        import ssl

        try:
            # Try to create an SSLContext, to be as rigorous as we can be
            # by avoiding making assumptions about the ApplicationRunner
            # implementation. If we happen to be on a Python that has no
            # SSLContext, we pass ssl=True, which will simply cause this
            # test to degenerate to the behavior of
            # test_conflict_SSL_True_with_ws_url (above). In fact, at the
            # moment (2015-05-10), none of this matters because the
            # ApplicationRunner implementation does not check to require
            # that its ssl argument is either a bool or an SSLContext. But
            # that may change, so we should be careful.
            ssl.create_default_context
        except AttributeError:
            context = True
        else:
            context = ssl.create_default_context()

        with replace_loop(Mock()) as loop:
            loop.run_until_complete = Mock(return_value=(Mock(), Mock()))
            runner = ApplicationRunner("ws://127.0.0.1:8080/wss", "realm", ssl=context)
            error = (
                r"^ssl argument value passed to ApplicationRunner "
                r'conflicts with the "ws:" prefix of the url '
                r'argument\. Did you mean to use "wss:"\?$'
            )
            self._assertRaisesRegex(Exception, error, runner.run, "_unused_")



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import unittest

from autobahn.rawsocket.util import create_url, parse_url


class TestCreateRsUrl(unittest.TestCase):
    def test_create_url01(self):
        self.assertEqual(create_url("localhost"), "rs://localhost:80")

    def test_create_url02(self):
        self.assertEqual(create_url("localhost", port=8090), "rs://localhost:8090")

    def test_create_url03(self):
        self.assertEqual(create_url("localhost", isSecure=True), "rss://localhost:443")

    def test_create_url04(self):
        self.assertEqual(
            create_url("localhost", isSecure=True, port=443), "rss://localhost:443"
        )

    def test_create_url05(self):
        self.assertEqual(
            create_url("localhost", isSecure=True, port=80), "rss://localhost:80"
        )

    def test_create_url06(self):
        self.assertEqual(create_url("unix", port="file.sock"), "rs://unix:file.sock")

    def test_create_url07(self):
        self.assertEqual(
            create_url("unix", port="/tmp/file.sock"), "rs://unix:/tmp/file.sock"
        )

    def test_create_url08(self):
        self.assertEqual(
            create_url("unix", port="../file.sock"), "rs://unix:../file.sock"
        )

    def test_create_url09(self):
        self.assertEqual(
            create_url("unix", isSecure=True, port="file.sock"), "rss://unix:file.sock"
        )

    def test_create_url10(self):
        self.assertEqual(
            create_url("unix", isSecure=True, port="/tmp/file.sock"),
            "rss://unix:/tmp/file.sock",
        )

    def test_create_url11(self):
        self.assertEqual(
            create_url("unix", isSecure=True, port="../file.sock"),
            "rss://unix:../file.sock",
        )


class TestParseWsUrl(unittest.TestCase):
    # parse_url -> (isSecure, host, port)

    def test_parse_url01(self):
        self.assertEqual(parse_url("rs://localhost"), (False, "localhost", 80))

    def test_parse_url02(self):
        self.assertEqual(parse_url("rss://localhost"), (True, "localhost", 443))

    def test_parse_url03(self):
        self.assertEqual(parse_url("rs://localhost:9000"), (False, "localhost", 9000))

    def test_parse_url04(self):
        self.assertEqual(parse_url("rss://localhost:9000"), (True, "localhost", 9000))

    def test_parse_url05(self):
        self.assertRaises(Exception, parse_url, "ws://localhost")

    def test_parse_url06(self):
        self.assertRaises(Exception, parse_url, "wss://localhost")

    def test_parse_url07(self):
        self.assertRaises(Exception, parse_url, "ws://localhost:80")

    def test_parse_url08(self):
        self.assertRaises(Exception, parse_url, "rs://localhost/somepath")

    def test_parse_url09(self):
        self.assertRaises(Exception, parse_url, "rs://localhost#somefrag")

    def test_parse_url10(self):
        self.assertRaises(Exception, parse_url, "rs://localhost?foo=bar")

    def test_parse_url11(self):
        self.assertRaises(Exception, parse_url, "rss://")

    def test_parse_url12(self):
        self.assertRaises(Exception, parse_url, "rs://")

    def test_parse_url13(self):
        self.assertEqual(parse_url("rs://unix:file.sock"), (False, "unix", "file.sock"))

    def test_parse_url14(self):
        self.assertEqual(
            parse_url("rs://unix:/tmp/file.sock"), (False, "unix", "/tmp/file.sock")
        )

    def test_parse_url15(self):
        self.assertEqual(
            parse_url("rs://unix:../file.sock"), (False, "unix", "../file.sock")
        )

    def test_parse_url16(self):
        self.assertEqual(parse_url("rss://unix:file.sock"), (True, "unix", "file.sock"))

    def test_parse_url17(self):
        self.assertEqual(
            parse_url("rss://unix:/tmp/file.sock"), (True, "unix", "/tmp/file.sock")
        )

    def test_parse_url18(self):
        self.assertEqual(
            parse_url("rss://unix:../file.sock"), (True, "unix", "../file.sock")
        )



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class Result(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Result()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsResult(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Result
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Result
    def Session(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # Result
    def Request(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # Positional values for application-defined call result.
    # Result
    def Args(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1),
            )
        return 0

    # Result
    def ArgsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Result
    def ArgsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Result
    def ArgsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        return o == 0

    # Keyword values for application-defined call result.
    # Result
    def Kwargs(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1),
            )
        return 0

    # Result
    def KwargsAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Result
    def KwargsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Result
    def KwargsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        return o == 0

    # Alternative, transparent payload. If given, ``args`` and ``kwargs`` must be left unset.
    # Result
    def Payload(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            a = self._tab.Vector(o)
            return self._tab.Get(
                flatbuffers.number_types.Uint8Flags,
                a + flatbuffers.number_types.UOffsetTFlags.py_type(j * 1),
            )
        return 0

    # Result
    def PayloadAsNumpy(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.GetVectorAsNumpy(flatbuffers.number_types.Uint8Flags, o)
        return 0

    # Result
    def PayloadLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Result
    def PayloadIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        return o == 0

    # The specific scheme in use with Payload Passthru (PPT) mode for the application payload.
    # Result
    def PptScheme(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # The specific serializer encoding the application payload with the Payload Passthru (PPT) scheme in use.
    # Result
    def PptSerializer(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # The cryptographic algorithm ("cipher") encrypting the application payload with the Payload Passthru (PPT) scheme in use.
    # Result
    def PptCipher(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # The identifier or reference to the encryption key that was used to encrypt the payload with the Payload Passthru (PPT) scheme and cipher in use.
    # Result
    def PptKeyid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Result
    def Progress(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # Result
    def Callee(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # Result
    def CalleeAuthid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Result
    def CalleeAuthrole(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Result
    def ForwardFor(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from autobahn.wamp.gen.wamp.proto.Principal import Principal

            obj = Principal()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Result
    def ForwardForLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Result
    def ForwardForIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        return o == 0


def ResultStart(builder):
    builder.StartObject(14)


def Start(builder):
    ResultStart(builder)


def ResultAddSession(builder, session):
    builder.PrependUint64Slot(0, session, 0)


def AddSession(builder, session):
    ResultAddSession(builder, session)


def ResultAddRequest(builder, request):
    builder.PrependUint64Slot(1, request, 0)


def AddRequest(builder, request):
    ResultAddRequest(builder, request)


def ResultAddArgs(builder, args):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(args), 0
    )


def AddArgs(builder, args):
    ResultAddArgs(builder, args)


def ResultStartArgsVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartArgsVector(builder, numElems):
    return ResultStartArgsVector(builder, numElems)


def ResultAddKwargs(builder, kwargs):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(kwargs), 0
    )


def AddKwargs(builder, kwargs):
    ResultAddKwargs(builder, kwargs)


def ResultStartKwargsVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartKwargsVector(builder, numElems):
    return ResultStartKwargsVector(builder, numElems)


def ResultAddPayload(builder, payload):
    builder.PrependUOffsetTRelativeSlot(
        4, flatbuffers.number_types.UOffsetTFlags.py_type(payload), 0
    )


def AddPayload(builder, payload):
    ResultAddPayload(builder, payload)


def ResultStartPayloadVector(builder, numElems):
    return builder.StartVector(1, numElems, 1)


def StartPayloadVector(builder, numElems):
    return ResultStartPayloadVector(builder, numElems)


def ResultAddPptScheme(builder, pptScheme):
    builder.PrependUint8Slot(5, pptScheme, 0)


def AddPptScheme(builder, pptScheme):
    ResultAddPptScheme(builder, pptScheme)


def ResultAddPptSerializer(builder, pptSerializer):
    builder.PrependUint8Slot(6, pptSerializer, 0)


def AddPptSerializer(builder, pptSerializer):
    ResultAddPptSerializer(builder, pptSerializer)


def ResultAddPptCipher(builder, pptCipher):
    builder.PrependUint8Slot(7, pptCipher, 0)


def AddPptCipher(builder, pptCipher):
    ResultAddPptCipher(builder, pptCipher)


def ResultAddPptKeyid(builder, pptKeyid):
    builder.PrependUOffsetTRelativeSlot(
        8, flatbuffers.number_types.UOffsetTFlags.py_type(pptKeyid), 0
    )


def AddPptKeyid(builder, pptKeyid):
    ResultAddPptKeyid(builder, pptKeyid)


def ResultAddProgress(builder, progress):
    builder.PrependBoolSlot(9, progress, 0)


def AddProgress(builder, progress):
    ResultAddProgress(builder, progress)


def ResultAddCallee(builder, callee):
    builder.PrependUint64Slot(10, callee, 0)


def AddCallee(builder, callee):
    ResultAddCallee(builder, callee)


def ResultAddCalleeAuthid(builder, calleeAuthid):
    builder.PrependUOffsetTRelativeSlot(
        11, flatbuffers.number_types.UOffsetTFlags.py_type(calleeAuthid), 0
    )


def AddCalleeAuthid(builder, calleeAuthid):
    ResultAddCalleeAuthid(builder, calleeAuthid)


def ResultAddCalleeAuthrole(builder, calleeAuthrole):
    builder.PrependUOffsetTRelativeSlot(
        12, flatbuffers.number_types.UOffsetTFlags.py_type(calleeAuthrole), 0
    )


def AddCalleeAuthrole(builder, calleeAuthrole):
    ResultAddCalleeAuthrole(builder, calleeAuthrole)


def ResultAddForwardFor(builder, forwardFor):
    builder.PrependUOffsetTRelativeSlot(
        13, flatbuffers.number_types.UOffsetTFlags.py_type(forwardFor), 0
    )


def AddForwardFor(builder, forwardFor):
    ResultAddForwardFor(builder, forwardFor)


def ResultStartForwardForVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartForwardForVector(builder, numElems):
    return ResultStartForwardForVector(builder, numElems)


def ResultEnd(builder):
    return builder.EndObject()


def End(builder):
    return ResultEnd(builder)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import asyncio
import hashlib
from asyncio import sleep  # noqa
from subprocess import Popen
from typing import Optional

from autobahn.wamp.types import TransportDetails

__all = (
    "sleep",
    "peer2str",
    "transport_channel_id",
    "create_transport_details",
)


def transport_channel_id(
    transport, is_server: bool, channel_id_type: Optional[str] = None
) -> bytes:
    """
    Application-layer user authentication protocols are vulnerable to generic
    credential forwarding attacks, where an authentication credential sent by
    a client C to a server M may then be used by M to impersonate C at another
    server S. To prevent such credential forwarding attacks, modern authentication
    protocols rely on channel bindings. For example, WAMP-cryptosign can use
    the tls-unique channel identifier provided by the TLS layer to strongly bind
    authentication credentials to the underlying channel, so that a credential
    received on one TLS channel cannot be forwarded on another.

    :param transport: The asyncio TLS transport to extract the TLS channel ID from.
    :param is_server: Flag indicating the transport is for a server.
    :param channel_id_type: TLS channel ID type, currently only "tls-unique" is supported.
    :returns: The TLS channel id (32 bytes).
    """
    if channel_id_type is None:
        return b"\x00" * 32

    # ssl.CHANNEL_BINDING_TYPES
    if channel_id_type not in ["tls-unique"]:
        raise Exception("invalid channel ID type {}".format(channel_id_type))

    ssl_obj = transport.get_extra_info("ssl_object")
    if ssl_obj is None:
        raise Exception(
            "TLS transport channel_id for tls-unique requested, but ssl_obj not found on transport"
        )

    if not hasattr(ssl_obj, "get_channel_binding"):
        raise Exception(
            "TLS transport channel_id for tls-unique requested, but get_channel_binding not found on ssl_obj"
        )

    # https://python.readthedocs.io/en/latest/library/ssl.html#ssl.SSLSocket.get_channel_binding
    # https://tools.ietf.org/html/rfc5929.html
    tls_finished_msg: bytes = ssl_obj.get_channel_binding(cb_type="tls-unique")

    if type(tls_finished_msg) != bytes:
        return b"\x00" * 32
    else:
        m = hashlib.sha256()
        m.update(tls_finished_msg)
        channel_id = m.digest()
        return channel_id


def peer2str(transport: asyncio.transports.BaseTransport) -> str:
    # https://docs.python.org/3.9/library/asyncio-protocol.html?highlight=get_extra_info#asyncio.BaseTransport.get_extra_info
    # https://docs.python.org/3.9/library/socket.html#socket.socket.getpeername
    try:
        peer = transport.get_extra_info("peername")
        if isinstance(peer, tuple):
            ip_ver = 4 if len(peer) == 2 else 6
            return "tcp{2}:{0}:{1}".format(peer[0], peer[1], ip_ver)
        elif isinstance(peer, str):
            return "unix:{0}".format(peer)
        else:
            return "?:{0}".format(peer)
    except:
        pass

    try:
        proc: Popen = transport.get_extra_info("subprocess")
        # return 'process:{}'.format(transport.pid)
        return "process:{}".format(proc.pid)
    except:
        pass

    try:
        pipe = transport.get_extra_info("pipe")
        return "pipe:{}".format(pipe)
    except:
        pass

    # gracefully fallback if we can't map the peer's transport
    return "unknown"


def get_serializers():
    from autobahn.wamp import serializer

    serializers = [
        "CBORSerializer",
        "MsgPackSerializer",
        "UBJSONSerializer",
        "JsonSerializer",
    ]
    serializers = list(
        filter(
            lambda x: x,
            map(
                lambda s: getattr(serializer, s) if hasattr(serializer, s) else None,
                serializers,
            ),
        )
    )
    return serializers


def create_transport_details(transport, is_server: bool) -> TransportDetails:
    # Internal helper. Base class calls this to create a TransportDetails
    peer = peer2str(transport)

    # https://docs.python.org/3.9/library/asyncio-protocol.html?highlight=get_extra_info#asyncio.BaseTransport.get_extra_info
    is_secure = transport.get_extra_info("peercert", None) is not None
    if is_secure:
        channel_id = {
            "tls-unique": transport_channel_id(transport, is_server, "tls-unique"),
        }
        channel_type = TransportDetails.CHANNEL_TYPE_TLS
        peer_cert = None
    else:
        channel_id = {}
        channel_type = TransportDetails.CHANNEL_TYPE_TCP
        peer_cert = None
    channel_framing = TransportDetails.CHANNEL_FRAMING_WEBSOCKET

    return TransportDetails(
        channel_type=channel_type,
        channel_framing=channel_framing,
        peer=peer,
        is_server=is_server,
        is_secure=is_secure,
        channel_id=channel_id,
        peer_cert=peer_cert,
    )



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from autobahn.wamp.exception import (
    ApplicationError,
    Error,
    InvalidUri,
    ProtocolError,
    SerializationError,
    SessionNotReady,
    TransportLost,
)
from autobahn.wamp.interfaces import ISession
from autobahn.wamp.types import (
    CallDetails,
    CallOptions,
    CallResult,
    CloseDetails,
    ComponentConfig,
    EventDetails,
    PublishOptions,
    RegisterOptions,
    SessionDetails,
    SubscribeOptions,
)
from autobahn.wamp.uri import error, register, subscribe

__all__ = (
    "ApplicationError",
    "CallDetails",
    "CallOptions",
    "CallResult",
    "CloseDetails",
    "ComponentConfig",
    "Error",
    "EventDetails",
    "ISession",
    "InvalidUri",
    "ProtocolError",
    "PublishOptions",
    "RegisterOptions",
    "SerializationError",
    "SessionDetails",
    "SessionNotReady",
    "SubscribeOptions",
    "TransportLost",
    "error",
    "register",
    "subscribe",
)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class Challenge(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Challenge()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsChallenge(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Challenge
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Challenge
    def Session(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # Challenge
    def Method(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0

    # Challenge
    def Extra(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from wamp.Map import Map

            obj = Map()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None


def ChallengeStart(builder):
    builder.StartObject(3)


def Start(builder):
    ChallengeStart(builder)


def ChallengeAddSession(builder, session):
    builder.PrependUint64Slot(0, session, 0)


def AddSession(builder, session):
    ChallengeAddSession(builder, session)


def ChallengeAddMethod(builder, method):
    builder.PrependUint8Slot(1, method, 0)


def AddMethod(builder, method):
    ChallengeAddMethod(builder, method)


def ChallengeAddExtra(builder, extra):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(extra), 0
    )


def AddExtra(builder, extra):
    ChallengeAddExtra(builder, extra)


def ChallengeEnd(builder):
    return builder.EndObject()


def End(builder):
    return ChallengeEnd(builder)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import unittest

from autobahn import wamp
from autobahn.wamp.uri import Pattern, RegisterOptions, SubscribeOptions


class TestUris(unittest.TestCase):
    def test_invalid_uris(self):
        for u in [
            "",
            "com.myapp.<product:foo>.update",
            "com.myapp.<123:int>.update",
            "com.myapp.<:product>.update",
            "com.myapp.<product:>.update",
            "com.myapp.<int:>.update",
        ]:
            self.assertRaises(Exception, Pattern, u, Pattern.URI_TARGET_ENDPOINT)

    def test_valid_uris(self):
        for u in [
            "com.myapp.proc1",
            "123",
            "com.myapp.<product:int>.update",
            "com.myapp.<category:string>.<subcategory>.listcom.myapp.something..update",
        ]:
            p = Pattern(u, Pattern.URI_TARGET_ENDPOINT)
            self.assertIsInstance(p, Pattern)

    def test_parse_uris(self):
        tests = [
            (
                "com.myapp.<product:int>.update",
                [
                    ("com.myapp.0.update", {"product": 0}),
                    ("com.myapp.123456.update", {"product": 123456}),
                    ("com.myapp.aaa.update", None),
                    ("com.myapp..update", None),
                    ("com.myapp.0.delete", None),
                ],
            ),
            (
                "com.myapp.<product:string>.update",
                [
                    ("com.myapp.box.update", {"product": "box"}),
                    ("com.myapp.123456.update", {"product": "123456"}),
                    ("com.myapp..update", None),
                ],
            ),
            (
                "com.myapp.<product>.update",
                [
                    ("com.myapp.0.update", {"product": "0"}),
                    ("com.myapp.abc.update", {"product": "abc"}),
                    ("com.myapp..update", None),
                ],
            ),
            (
                "com.myapp.<category:string>.<subcategory:string>.list",
                [
                    (
                        "com.myapp.cosmetic.shampoo.list",
                        {"category": "cosmetic", "subcategory": "shampoo"},
                    ),
                    ("com.myapp...list", None),
                    ("com.myapp.cosmetic..list", None),
                    ("com.myapp..shampoo.list", None),
                ],
            ),
            (
                "eth.pydefi.tradeclock.<clock_oid:str>.get_clock_info",
                [
                    (
                        "eth.pydefi.tradeclock.ba3b1e9f-3006-4eae-ae88-cf5896b36342.get_clock_info",
                        {"clock_oid": "ba3b1e9f-3006-4eae-ae88-cf5896b36342"},
                    ),
                ],
            ),
            (
                "eth.wamp.network.catalog.<catalog_adr:str>.owner",
                [
                    (
                        "eth.wamp.network.catalog.0xAA8Cc377db31a354137d8Bb86D0E38495dbD5266.owner",
                        {"catalog_adr": "0xAA8Cc377db31a354137d8Bb86D0E38495dbD5266"},
                    ),
                ],
            ),
        ]
        for test in tests:
            pat = Pattern(test[0], Pattern.URI_TARGET_ENDPOINT)
            for ptest in test[1]:
                uri = ptest[0]
                kwargs_should = ptest[1]
                if kwargs_should is not None:
                    args_is, kwargs_is = pat.match(uri)
                    self.assertEqual(kwargs_is, kwargs_should)
                else:
                    self.assertRaises(Exception, pat.match, uri)


class TestDecorators(unittest.TestCase):
    def test_decorate_endpoint(self):
        @wamp.register("com.calculator.square")
        def square(_):
            """Do nothing."""

        self.assertTrue(hasattr(square, "_wampuris"))
        self.assertTrue(type(square._wampuris) == list)
        self.assertEqual(len(square._wampuris), 1)
        self.assertIsInstance(square._wampuris[0], Pattern)
        self.assertTrue(square._wampuris[0].is_endpoint())
        self.assertFalse(square._wampuris[0].is_handler())
        self.assertFalse(square._wampuris[0].is_exception())
        self.assertEqual(square._wampuris[0].uri(), "com.calculator.square")
        self.assertEqual(square._wampuris[0]._type, Pattern.URI_TYPE_EXACT)

        @wamp.register("com.myapp.product.<product:int>.update")
        def update_product(product=None, label=None):
            """Do nothing."""

        self.assertTrue(hasattr(update_product, "_wampuris"))
        self.assertTrue(type(update_product._wampuris) == list)
        self.assertEqual(len(update_product._wampuris), 1)
        self.assertIsInstance(update_product._wampuris[0], Pattern)
        self.assertTrue(update_product._wampuris[0].is_endpoint())
        self.assertFalse(update_product._wampuris[0].is_handler())
        self.assertFalse(update_product._wampuris[0].is_exception())
        self.assertEqual(
            update_product._wampuris[0].uri(), "com.myapp.product.<product:int>.update"
        )
        self.assertEqual(update_product._wampuris[0]._type, Pattern.URI_TYPE_WILDCARD)

        @wamp.register("com.myapp.<category:string>.<cid:int>.update")
        def update(category=None, cid=None):
            """Do nothing."""

        self.assertTrue(hasattr(update, "_wampuris"))
        self.assertTrue(type(update._wampuris) == list)
        self.assertEqual(len(update._wampuris), 1)
        self.assertIsInstance(update._wampuris[0], Pattern)
        self.assertTrue(update._wampuris[0].is_endpoint())
        self.assertFalse(update._wampuris[0].is_handler())
        self.assertFalse(update._wampuris[0].is_exception())
        self.assertEqual(
            update._wampuris[0].uri(), "com.myapp.<category:string>.<cid:int>.update"
        )
        self.assertEqual(update._wampuris[0]._type, Pattern.URI_TYPE_WILDCARD)

        @wamp.register(
            "com.myapp.circle.<name:string>",
            RegisterOptions(match="wildcard", details_arg="details"),
        )
        def circle(name=None, details=None):
            """Do nothing."""

        self.assertTrue(hasattr(circle, "_wampuris"))
        self.assertTrue(type(circle._wampuris) == list)
        self.assertEqual(len(circle._wampuris), 1)
        self.assertIsInstance(circle._wampuris[0], Pattern)
        self.assertIsInstance(circle._wampuris[0].options, RegisterOptions)
        self.assertEqual(circle._wampuris[0].options.match, "wildcard")
        self.assertEqual(circle._wampuris[0].options.details_arg, "details")
        self.assertTrue(circle._wampuris[0].is_endpoint())
        self.assertFalse(circle._wampuris[0].is_handler())
        self.assertFalse(circle._wampuris[0].is_exception())
        self.assertEqual(circle._wampuris[0].uri(), "com.myapp.circle.<name:string>")
        self.assertEqual(circle._wampuris[0]._type, Pattern.URI_TYPE_WILDCARD)

        @wamp.register(
            "com.myapp.something..update",
            RegisterOptions(match="wildcard", details_arg="details"),
        )
        def something(dynamic=None, details=None):
            """Do nothing."""

        self.assertTrue(hasattr(something, "_wampuris"))
        self.assertTrue(type(something._wampuris) == list)
        self.assertEqual(len(something._wampuris), 1)
        self.assertIsInstance(something._wampuris[0], Pattern)
        self.assertIsInstance(something._wampuris[0].options, RegisterOptions)
        self.assertEqual(something._wampuris[0].options.match, "wildcard")
        self.assertEqual(something._wampuris[0].options.details_arg, "details")
        self.assertTrue(something._wampuris[0].is_endpoint())
        self.assertFalse(something._wampuris[0].is_handler())
        self.assertFalse(something._wampuris[0].is_exception())
        self.assertEqual(something._wampuris[0].uri(), "com.myapp.something..update")
        self.assertEqual(something._wampuris[0]._type, Pattern.URI_TYPE_WILDCARD)

    def test_decorate_handler(self):
        @wamp.subscribe("com.myapp.on_shutdown")
        def on_shutdown():
            """Do nothing."""

        self.assertTrue(hasattr(on_shutdown, "_wampuris"))
        self.assertTrue(type(on_shutdown._wampuris) == list)
        self.assertEqual(len(on_shutdown._wampuris), 1)
        self.assertIsInstance(on_shutdown._wampuris[0], Pattern)
        self.assertFalse(on_shutdown._wampuris[0].is_endpoint())
        self.assertTrue(on_shutdown._wampuris[0].is_handler())
        self.assertFalse(on_shutdown._wampuris[0].is_exception())
        self.assertEqual(on_shutdown._wampuris[0].uri(), "com.myapp.on_shutdown")
        self.assertEqual(on_shutdown._wampuris[0]._type, Pattern.URI_TYPE_EXACT)

        @wamp.subscribe("com.myapp.product.<product:int>.on_update")
        def on_product_update(product=None, label=None):
            """Do nothing."""

        self.assertTrue(hasattr(on_product_update, "_wampuris"))
        self.assertTrue(type(on_product_update._wampuris) == list)
        self.assertEqual(len(on_product_update._wampuris), 1)
        self.assertIsInstance(on_product_update._wampuris[0], Pattern)
        self.assertFalse(on_product_update._wampuris[0].is_endpoint())
        self.assertTrue(on_product_update._wampuris[0].is_handler())
        self.assertFalse(on_product_update._wampuris[0].is_exception())
        self.assertEqual(
            on_product_update._wampuris[0].uri(),
            "com.myapp.product.<product:int>.on_update",
        )
        self.assertEqual(
            on_product_update._wampuris[0]._type, Pattern.URI_TYPE_WILDCARD
        )

        @wamp.subscribe("com.myapp.<category:string>.<cid:int>.on_update")
        def on_update(category=None, cid=None, label=None):
            """Do nothing."""

        self.assertTrue(hasattr(on_update, "_wampuris"))
        self.assertTrue(type(on_update._wampuris) == list)
        self.assertEqual(len(on_update._wampuris), 1)
        self.assertIsInstance(on_update._wampuris[0], Pattern)
        self.assertFalse(on_update._wampuris[0].is_endpoint())
        self.assertTrue(on_update._wampuris[0].is_handler())
        self.assertFalse(on_update._wampuris[0].is_exception())
        self.assertEqual(
            on_update._wampuris[0].uri(),
            "com.myapp.<category:string>.<cid:int>.on_update",
        )
        self.assertEqual(on_update._wampuris[0]._type, Pattern.URI_TYPE_WILDCARD)

        @wamp.subscribe(
            "com.myapp.on.<event:string>",
            SubscribeOptions(match="wildcard", details_arg="details"),
        )
        def on_event(event=None, details=None):
            """Do nothing."""

        self.assertTrue(hasattr(on_event, "_wampuris"))
        self.assertTrue(type(on_event._wampuris) == list)
        self.assertEqual(len(on_event._wampuris), 1)
        self.assertIsInstance(on_event._wampuris[0], Pattern)
        self.assertIsInstance(on_event._wampuris[0].options, SubscribeOptions)
        self.assertEqual(on_event._wampuris[0].options.match, "wildcard")
        self.assertEqual(on_event._wampuris[0].options.details_arg, "details")
        self.assertFalse(on_event._wampuris[0].is_endpoint())
        self.assertTrue(on_event._wampuris[0].is_handler())
        self.assertFalse(on_event._wampuris[0].is_exception())
        self.assertEqual(on_event._wampuris[0].uri(), "com.myapp.on.<event:string>")
        self.assertEqual(on_event._wampuris[0]._type, Pattern.URI_TYPE_WILDCARD)

    def test_decorate_exception(self):
        @wamp.error("com.myapp.error")
        class AppError(Exception):
            """Do nothing."""

        self.assertTrue(hasattr(AppError, "_wampuris"))
        self.assertTrue(type(AppError._wampuris) == list)
        self.assertEqual(len(AppError._wampuris), 1)
        self.assertIsInstance(AppError._wampuris[0], Pattern)
        self.assertFalse(AppError._wampuris[0].is_endpoint())
        self.assertFalse(AppError._wampuris[0].is_handler())
        self.assertTrue(AppError._wampuris[0].is_exception())
        self.assertEqual(AppError._wampuris[0].uri(), "com.myapp.error")
        self.assertEqual(AppError._wampuris[0]._type, Pattern.URI_TYPE_EXACT)

        @wamp.error("com.myapp.product.<product:int>.product_inactive")
        class ProductInactiveError(Exception):
            """Do nothing."""

        self.assertTrue(hasattr(ProductInactiveError, "_wampuris"))
        self.assertTrue(type(ProductInactiveError._wampuris) == list)
        self.assertEqual(len(ProductInactiveError._wampuris), 1)
        self.assertIsInstance(ProductInactiveError._wampuris[0], Pattern)
        self.assertFalse(ProductInactiveError._wampuris[0].is_endpoint())
        self.assertFalse(ProductInactiveError._wampuris[0].is_handler())
        self.assertTrue(ProductInactiveError._wampuris[0].is_exception())
        self.assertEqual(
            ProductInactiveError._wampuris[0].uri(),
            "com.myapp.product.<product:int>.product_inactive",
        )
        self.assertEqual(
            ProductInactiveError._wampuris[0]._type, Pattern.URI_TYPE_WILDCARD
        )

        @wamp.error("com.myapp.<category:string>.<product:int>.inactive")
        class ObjectInactiveError(Exception):
            """Do nothing."""

        self.assertTrue(hasattr(ObjectInactiveError, "_wampuris"))
        self.assertTrue(type(ObjectInactiveError._wampuris) == list)
        self.assertEqual(len(ObjectInactiveError._wampuris), 1)
        self.assertIsInstance(ObjectInactiveError._wampuris[0], Pattern)
        self.assertFalse(ObjectInactiveError._wampuris[0].is_endpoint())
        self.assertFalse(ObjectInactiveError._wampuris[0].is_handler())
        self.assertTrue(ObjectInactiveError._wampuris[0].is_exception())
        self.assertEqual(
            ObjectInactiveError._wampuris[0].uri(),
            "com.myapp.<category:string>.<product:int>.inactive",
        )
        self.assertEqual(
            ObjectInactiveError._wampuris[0]._type, Pattern.URI_TYPE_WILDCARD
        )

    def test_match_decorated_endpoint(self):
        @wamp.register("com.calculator.square")
        def square(x):
            return x

        args, kwargs = square._wampuris[0].match("com.calculator.square")
        self.assertEqual(square(666, **kwargs), 666)

        @wamp.register("com.myapp.product.<product:int>.update")
        def update_product(product=None, label=None):
            return product, label

        args, kwargs = update_product._wampuris[0].match(
            "com.myapp.product.123456.update"
        )
        kwargs["label"] = "foobar"
        self.assertEqual(update_product(**kwargs), (123456, "foobar"))

        @wamp.register("com.myapp.<category:string>.<cid:int>.update")
        def update(category=None, cid=None, label=None):
            return category, cid, label

        args, kwargs = update._wampuris[0].match("com.myapp.product.123456.update")
        kwargs["label"] = "foobar"
        self.assertEqual(update(**kwargs), ("product", 123456, "foobar"))

    def test_match_decorated_handler(self):
        @wamp.subscribe("com.myapp.on_shutdown")
        def on_shutdown():
            pass

        args, kwargs = on_shutdown._wampuris[0].match("com.myapp.on_shutdown")
        self.assertEqual(on_shutdown(**kwargs), None)

        @wamp.subscribe("com.myapp.product.<product:int>.on_update")
        def on_product_update(product=None, label=None):
            return product, label

        args, kwargs = on_product_update._wampuris[0].match(
            "com.myapp.product.123456.on_update"
        )
        kwargs["label"] = "foobar"
        self.assertEqual(on_product_update(**kwargs), (123456, "foobar"))

        @wamp.subscribe("com.myapp.<category:string>.<cid:int>.on_update")
        def on_update(category=None, cid=None, label=None):
            return category, cid, label

        args, kwargs = on_update._wampuris[0].match(
            "com.myapp.product.123456.on_update"
        )
        kwargs["label"] = "foobar"
        self.assertEqual(on_update(**kwargs), ("product", 123456, "foobar"))

    def test_match_decorated_exception(self):
        @wamp.error("com.myapp.error")
        class AppError(Exception):
            def __init__(self, msg):
                Exception.__init__(self, msg)

            def __eq__(self, other):
                return self.__class__ == other.__class__ and self.args == other.args

        args, kwargs = AppError._wampuris[0].match("com.myapp.error")
        # noinspection PyArgumentList
        self.assertEqual(AppError("fuck", **kwargs), AppError("fuck"))

        @wamp.error("com.myapp.product.<product:int>.product_inactive")
        class ProductInactiveError(Exception):
            def __init__(self, msg, product=None):
                Exception.__init__(self, msg)
                self.product = product

            def __eq__(self, other):
                return (
                    self.__class__ == other.__class__
                    and self.args == other.args
                    and self.product == other.product
                )

        args, kwargs = ProductInactiveError._wampuris[0].match(
            "com.myapp.product.123456.product_inactive"
        )
        self.assertEqual(
            ProductInactiveError("fuck", **kwargs), ProductInactiveError("fuck", 123456)
        )

        @wamp.error("com.myapp.<category:string>.<product:int>.inactive")
        class ObjectInactiveError(Exception):
            def __init__(self, msg, category=None, product=None):
                Exception.__init__(self, msg)
                self.category = category
                self.product = product

            def __eq__(self, other):
                return (
                    self.__class__ == other.__class__
                    and self.args == other.args
                    and self.category == other.category
                    and self.product == other.product
                )

        args, kwargs = ObjectInactiveError._wampuris[0].match(
            "com.myapp.product.123456.inactive"
        )
        self.assertEqual(
            ObjectInactiveError("fuck", **kwargs),
            ObjectInactiveError("fuck", "product", 123456),
        )


class KwException(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args)
        self.kwargs = kwargs


# what if the WAMP error message received
# contains args/kwargs that cannot be
# consumed by the constructor of the exception
# class defined for the WAMP error URI?

# 1. we can bail out (but we are already signaling an error)
# 2. we can require a generic constructor
# 3. we can map only unconsumed args/kwargs to generic attributes
# 4. we can silently drop unconsumed args/kwargs


class MockSession(object):
    def __init__(self):
        self._ecls_to_uri_pat = {}
        self._uri_to_ecls = {}

    def define(self, exception, error=None):
        if error is None:
            assert hasattr(exception, "_wampuris")
            self._ecls_to_uri_pat[exception] = exception._wampuris
            self._uri_to_ecls[exception._wampuris[0].uri()] = exception
        else:
            assert not hasattr(exception, "_wampuris")
            self._ecls_to_uri_pat[exception] = [
                Pattern(error, Pattern.URI_TARGET_HANDLER)
            ]
            self._uri_to_ecls[error] = exception

    def map_error(self, error, args=None, kwargs=None):
        # FIXME:
        # 1. map to ecls based on error URI wildcard/prefix
        # 2. extract additional args/kwargs from error URI

        if error in self._uri_to_ecls:
            ecls = self._uri_to_ecls[error]
            try:
                # the following might fail, eg. TypeError when
                # signature of exception constructor is incompatible
                # with args/kwargs or when the exception constructor raises
                if kwargs:
                    if args:
                        exc = ecls(*args, **kwargs)
                    else:
                        exc = ecls(**kwargs)
                else:
                    if args:
                        exc = ecls(*args)
                    else:
                        exc = ecls()
            except Exception:
                # FIXME: log e
                exc = KwException(error, *args, **kwargs)
        else:
            # this never fails
            args = args or []
            kwargs = kwargs or {}
            exc = KwException(error, *args, **kwargs)
        return exc


class TestDecoratorsAdvanced(unittest.TestCase):
    def test_decorate_exception_non_exception(self):
        def test():
            # noinspection PyUnusedLocal
            @wamp.error("com.test.error")
            class Foo(object):
                pass

        self.assertRaises(Exception, test)

    def test_decorate_endpoint_multiple(self):
        # noinspection PyUnusedLocal
        @wamp.register("com.oldapp.oldproc")
        @wamp.register("com.calculator.square")
        def square(x):
            """Do nothing."""

        self.assertTrue(hasattr(square, "_wampuris"))
        self.assertTrue(type(square._wampuris) == list)
        self.assertEqual(len(square._wampuris), 2)

        for i in range(2):
            self.assertIsInstance(square._wampuris[i], Pattern)
            self.assertTrue(square._wampuris[i].is_endpoint())
            self.assertFalse(square._wampuris[i].is_handler())
            self.assertFalse(square._wampuris[i].is_exception())
            self.assertEqual(square._wampuris[i]._type, Pattern.URI_TYPE_EXACT)

        self.assertEqual(square._wampuris[0].uri(), "com.calculator.square")
        self.assertEqual(square._wampuris[1].uri(), "com.oldapp.oldproc")

    def test_marshal_decorated_exception(self):
        @wamp.error("com.myapp.error")
        class AppError(Exception):
            pass

        try:
            raise AppError("fuck")
        except Exception as e:
            self.assertEqual(e._wampuris[0].uri(), "com.myapp.error")

        @wamp.error("com.myapp.product.<product:int>.product_inactive")
        class ProductInactiveError(Exception):
            def __init__(self, msg, product=None):
                Exception.__init__(self, msg)
                self.product = product

        try:
            raise ProductInactiveError("fuck", 123456)
        except Exception as e:
            self.assertEqual(
                e._wampuris[0].uri(), "com.myapp.product.<product:int>.product_inactive"
            )

        session = MockSession()
        session.define(AppError)

    def test_define_exception_undecorated(self):
        session = MockSession()

        class AppError(Exception):
            pass

        # defining an undecorated exception requires
        # an URI to be provided
        self.assertRaises(Exception, session.define, AppError)

        session.define(AppError, "com.myapp.error")

        exc = session.map_error("com.myapp.error")
        self.assertIsInstance(exc, AppError)

    def test_define_exception_decorated(self):
        session = MockSession()

        @wamp.error("com.myapp.error")
        class AppError(Exception):
            pass

        # when defining a decorated exception
        # an URI must not be provided
        self.assertRaises(Exception, session.define, AppError, "com.myapp.error")

        session.define(AppError)

        exc = session.map_error("com.myapp.error")
        self.assertIsInstance(exc, AppError)

    def test_map_exception_undefined(self):
        session = MockSession()

        exc = session.map_error("com.myapp.error")
        self.assertIsInstance(exc, Exception)

    def test_map_exception_args(self):
        session = MockSession()

        @wamp.error("com.myapp.error")
        class AppError(Exception):
            pass

        @wamp.error("com.myapp.error.product_inactive")
        class ProductInactiveError(Exception):
            def __init__(self, product=None):
                self.product = product

        # define exceptions in mock session
        session.define(AppError)
        session.define(ProductInactiveError)

        for test in [
            # ("com.myapp.foo.error", [], {}, KwException),
            ("com.myapp.error", [], {}, AppError),
            ("com.myapp.error", ["you are doing it wrong"], {}, AppError),
            ("com.myapp.error", ["you are doing it wrong", 1, 2, 3], {}, AppError),
            ("com.myapp.error.product_inactive", [], {}, ProductInactiveError),
            (
                "com.myapp.error.product_inactive",
                [],
                {"product": 123456},
                ProductInactiveError,
            ),
        ]:
            error, args, kwargs, ecls = test
            exc = session.map_error(error, args, kwargs)

            self.assertIsInstance(exc, ecls)
            self.assertEqual(list(exc.args), args)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import os
import unittest.mock as mock

if os.environ.get("USE_TWISTED", False):
    import twisted
    from autobahn import util
    from autobahn.twisted.wamp import ApplicationSession, Session
    from autobahn.wamp import CloseDetails, message, role, serializer, types, uri
    from autobahn.wamp.auth import create_authenticator
    from autobahn.wamp.exception import (
        ApplicationError,
        InvalidUri,
        NotAuthorized,
        ProtocolError,
    )
    from autobahn.wamp.interfaces import IAuthenticator
    from autobahn.wamp.request import CallRequest
    from autobahn.wamp.types import TransportDetails
    from twisted.internet.defer import (
        Deferred,
        DeferredList,
        fail,
        inlineCallbacks,
        succeed,
    )
    from twisted.trial import unittest

    class MockTransport(object):
        def __init__(self, handler):
            self._log = False
            self._handler = handler
            self._serializer = serializer.JsonSerializer()
            self._registrations = {}
            self._invocations = {}
            #: str -> ID
            self._subscription_topics = {}

            self._handler.onOpen(self)

            self._my_session_id = util.id()

            roles = {
                "broker": role.RoleBrokerFeatures(),
                "dealer": role.RoleDealerFeatures(),
            }

            msg = message.Welcome(self._my_session_id, roles)
            self._handler.onMessage(msg)
            self._fake_router_session = ApplicationSession()

            self._transport_details = TransportDetails()

        def transport_details(self):
            return self._transport_details

        def drop_registration(self, reg_id):
            self._handler.onMessage(message.Unregistered(0, reg_id))

        def send(self, msg):
            if self._log:
                payload, isbinary = self._serializer.serialize(msg)
                print("Send: {0}".format(payload))

            reply = None

            if isinstance(msg, message.Publish):
                if msg.topic.startswith("com.myapp"):
                    if msg.acknowledge:
                        reply = message.Published(
                            msg.request,
                            self._fake_router_session._request_id_gen.next(),
                        )
                elif len(msg.topic) == 0:
                    reply = message.Error(
                        message.Publish.MESSAGE_TYPE,
                        msg.request,
                        "wamp.error.invalid_uri",
                    )
                elif msg.topic.startswith("noreply."):
                    pass
                else:
                    reply = message.Error(
                        message.Publish.MESSAGE_TYPE,
                        msg.request,
                        "wamp.error.not_authorized",
                    )

            elif isinstance(msg, message.Call):
                if msg.procedure == "com.myapp.procedure1":
                    reply = message.Result(msg.request, args=[100])
                elif msg.procedure == "com.myapp.procedure2":
                    reply = message.Result(msg.request, args=[1, 2, 3])
                elif msg.procedure == "com.myapp.procedure3":
                    reply = message.Result(
                        msg.request, args=[1, 2, 3], kwargs={"foo": "bar", "baz": 23}
                    )

                elif msg.procedure.startswith("com.myapp.myproc"):
                    registration = self._registrations[msg.procedure]
                    request = self._fake_router_session._request_id_gen.next()
                    if request in self._invocations:
                        raise ProtocolError("duplicate invocation")
                    self._invocations[request] = msg.request
                    reply = message.Invocation(
                        request,
                        registration,
                        args=msg.args,
                        kwargs=msg.kwargs,
                        receive_progress=msg.receive_progress,
                    )
                else:
                    reply = message.Error(
                        message.Call.MESSAGE_TYPE,
                        msg.request,
                        "wamp.error.no_such_procedure",
                    )

            elif isinstance(msg, message.Yield):
                if msg.request in self._invocations:
                    request = self._invocations[msg.request]
                    reply = message.Result(
                        request, args=msg.args, kwargs=msg.kwargs, progress=msg.progress
                    )

            elif isinstance(msg, message.Subscribe):
                topic = msg.topic
                if topic in self._subscription_topics:
                    reply_id = self._subscription_topics[topic]
                else:
                    reply_id = self._fake_router_session._request_id_gen.next()
                    self._subscription_topics[topic] = reply_id
                reply = message.Subscribed(msg.request, reply_id)

            elif isinstance(msg, message.Unsubscribe):
                reply = message.Unsubscribed(msg.request)

            elif isinstance(msg, message.Register):
                registration = self._fake_router_session._request_id_gen.next()
                self._registrations[msg.procedure] = registration
                reply = message.Registered(msg.request, registration)

            elif isinstance(msg, message.Unregister):
                reply = message.Unregistered(msg.request)

            elif isinstance(msg, message.Error):
                # since I'm basically a Dealer, I forward on the
                # error, but reply to my own request/invocation
                request = self._invocations[msg.request]
                reply = message.Error(
                    message.Call.MESSAGE_TYPE,
                    request,
                    msg.error,
                    args=msg.args,
                    kwargs=msg.kwargs,
                )

            if reply:
                if self._log:
                    payload, isbinary = self._serializer.serialize(reply)
                    print("Receive: {0}".format(payload))
                self._handler.onMessage(reply)

        def isOpen(self):
            return True

        def close(self):
            pass

        def abort(self):
            pass

    class TestClose(unittest.TestCase):
        def test_server_abort(self):
            handler = ApplicationSession()
            MockTransport(handler)

            # this should not raise an exception, but did when this
            # test-case was written
            handler.onClose(False)

        def test_reject_pending(self):
            handler = ApplicationSession()
            MockTransport(handler)

            # This could happen if the task waiting on a request gets cancelled
            deferred = fail(Exception())
            handler._call_reqs[1] = CallRequest(1, "foo", deferred, {})
            handler.onLeave(CloseDetails())

    class TestRegisterDecorator(unittest.TestCase):
        def test_prefix(self):
            class Prefix(ApplicationSession):
                @uri.register("method_name")
                def some_method(self):
                    pass

            session = Prefix()

            session._transport = mock.Mock()
            session.register(session, prefix="com.example.prefix.")

            # we should have registered one method, with the prefix
            # put in front
            self.assertEqual(1, len(session._transport.mock_calls))
            call = session._transport.mock_calls[0]
            self.assertEqual("send", call[0])
            reg = call.call_list()[0][1][0]
            self.assertEqual("com.example.prefix.method_name", reg.procedure)

        def test_auto_name(self):
            class Magic(ApplicationSession):
                @uri.register(None)
                def some_method(self):
                    pass

            session = Magic()

            session._transport = mock.Mock()
            session.register(session, prefix="com.example.")

            # Should auto-discover name by passing uri=None above
            self.assertEqual(1, len(session._transport.mock_calls))
            call = session._transport.mock_calls[0]
            self.assertEqual("send", call[0])
            reg = call.call_list()[0][1][0]
            self.assertEqual("com.example.some_method", reg.procedure)

    class TestPublisher(unittest.TestCase):
        @inlineCallbacks
        def test_publish(self):
            handler = ApplicationSession()
            MockTransport(handler)

            publication = yield handler.publish("com.myapp.topic1")
            self.assertEqual(publication, None)

            publication = yield handler.publish("com.myapp.topic1", 1, 2, 3)
            self.assertEqual(publication, None)

            publication = yield handler.publish(
                "com.myapp.topic1", 1, 2, 3, foo=23, bar="hello"
            )
            self.assertEqual(publication, None)

            publication = yield handler.publish(
                "com.myapp.topic1", options=types.PublishOptions(exclude_me=False)
            )
            self.assertEqual(publication, None)

            publication = yield handler.publish(
                "com.myapp.topic1",
                1,
                2,
                3,
                foo=23,
                bar="hello",
                options=types.PublishOptions(exclude_me=False, exclude=[100, 200, 300]),
            )
            self.assertEqual(publication, None)

            publication = yield handler.publish(
                "com.myapp.topic1",
                1,
                2,
                3,
                foo=23,
                bar="hello",
                options=types.PublishOptions(
                    exclude_me=False, exclude=[100, 200, 300], retain=True
                ),
            )
            self.assertEqual(publication, None)

        @inlineCallbacks
        def test_publish_outstanding_errors(self):
            handler = ApplicationSession()
            MockTransport(handler)

            # this publish will "hang" because 'noreply.' URI is
            # handled specially in MockTransport; so this request will
            # be outstanding
            publication = handler.publish(
                "noreply.foo",
                options=types.PublishOptions(acknowledge=True),
            )
            # "leave" the session, which should trigger errbacks on
            # all outstanding requests.
            details = types.CloseDetails(reason="testing", message="how are you?")
            yield handler.onLeave(details)

            # ensure we got our errback
            try:
                yield publication
            except ApplicationError as e:
                self.assertEqual("testing", e.error)
                self.assertEqual("how are you?", e.message)

        @inlineCallbacks
        def test_publish_outstanding_errors_async_errback(self):
            handler = ApplicationSession()
            MockTransport(handler)
            error_d = Deferred()

            # this publish will "hang" because 'noreply.' URI is
            # handled specially in MockTransport; so this request will
            # be outstanding
            publication_d = handler.publish(
                "noreply.foo",
                options=types.PublishOptions(acknowledge=True),
            )
            # further, we add an errback that does some arbitrary async work
            got_errors = []

            def errback(fail):
                got_errors.append(fail)
                return error_d

            publication_d.addErrback(errback)
            # "leave" the session, which should trigger errbacks on
            # all outstanding requests.
            details = types.CloseDetails(reason="testing", message="how are you?")
            handler.onLeave(details)

            # since our errback is async, onLeave should not have
            # completed yet but we should have already failed the
            # publication
            self.assertEqual(1, len(got_errors))
            # ...now let the async errback continue by completing the
            # Deferred we returned in our errback (could be fail or
            # success, shoudln't matter)
            error_d.callback(None)

            # ensure we (now) get our errback
            try:
                yield publication_d
            except ApplicationError as e:
                self.assertEqual("testing", e.error)
                self.assertEqual("how are you?", e.message)

        @inlineCallbacks
        def test_publish_acknowledged(self):
            handler = ApplicationSession()
            MockTransport(handler)

            publication = yield handler.publish(
                "com.myapp.topic1", options=types.PublishOptions(acknowledge=True)
            )
            self.assertTrue(type(publication.id) == int)

            publication = yield handler.publish(
                "com.myapp.topic1",
                1,
                2,
                3,
                options=types.PublishOptions(acknowledge=True),
            )
            self.assertTrue(type(publication.id) == int)

            publication = yield handler.publish(
                "com.myapp.topic1",
                1,
                2,
                3,
                foo=23,
                bar="hello",
                options=types.PublishOptions(acknowledge=True),
            )
            self.assertTrue(type(publication.id) == int)

            publication = yield handler.publish(
                "com.myapp.topic1",
                options=types.PublishOptions(exclude_me=False, acknowledge=True),
            )
            self.assertTrue(type(publication.id) == int)

            publication = yield handler.publish(
                "com.myapp.topic1",
                1,
                2,
                3,
                foo=23,
                bar="hello",
                options=types.PublishOptions(
                    exclude_me=False, exclude=[100, 200, 300], acknowledge=True
                ),
            )
            self.assertTrue(type(publication.id) == int)

        @inlineCallbacks
        def test_publish_undefined_exception(self):
            handler = ApplicationSession()
            MockTransport(handler)

            options = types.PublishOptions(acknowledge=True)

            yield self.assertFailure(
                handler.publish("de.myapp.topic1", options=options), ApplicationError
            )
            yield self.assertFailure(
                handler.publish("foobar", options=options), ApplicationError
            )

        @inlineCallbacks
        def test_publish_defined_exception(self):
            handler = ApplicationSession()
            MockTransport(handler)

            options = types.PublishOptions(acknowledge=True)

            handler.define(NotAuthorized)
            yield self.assertFailure(
                handler.publish("de.myapp.topic1", options=options), NotAuthorized
            )

            handler.define(InvalidUri)
            yield self.assertFailure(
                handler.publish("foobar", options=options), NotAuthorized
            )

        @inlineCallbacks
        def test_call(self):
            handler = ApplicationSession()
            MockTransport(handler)

            res = yield handler.call("com.myapp.procedure1")
            self.assertEqual(res, 100)

            res = yield handler.call("com.myapp.procedure1", 1, 2, 3)
            self.assertEqual(res, 100)

            res = yield handler.call(
                "com.myapp.procedure1", 1, 2, 3, foo=23, bar="hello"
            )
            self.assertEqual(res, 100)

            res = yield handler.call(
                "com.myapp.procedure1", options=types.CallOptions(timeout=10000)
            )
            self.assertEqual(res, 100)

            res = yield handler.call(
                "com.myapp.procedure1",
                1,
                2,
                3,
                foo=23,
                bar="hello",
                options=types.CallOptions(timeout=10000),
            )
            self.assertEqual(res, 100)

        @inlineCallbacks
        def test_call_with_complex_result(self):
            handler = ApplicationSession()
            MockTransport(handler)

            res = yield handler.call("com.myapp.procedure2")
            self.assertIsInstance(res, types.CallResult)
            self.assertEqual(res.results, (1, 2, 3))
            self.assertEqual(res.kwresults, {})

            res = yield handler.call("com.myapp.procedure3")
            self.assertIsInstance(res, types.CallResult)
            self.assertEqual(res.results, (1, 2, 3))
            self.assertEqual(res.kwresults, {"foo": "bar", "baz": 23})

        @inlineCallbacks
        def test_subscribe(self):
            handler = ApplicationSession()
            MockTransport(handler)

            def on_event(*args, **kwargs):
                print("got event", args, kwargs)

            subscription = yield handler.subscribe(on_event, "com.myapp.topic1")
            self.assertTrue(type(subscription.id) == int)

            subscription = yield handler.subscribe(
                on_event,
                "com.myapp.topic1",
                options=types.SubscribeOptions(match="wildcard"),
            )
            self.assertTrue(type(subscription.id) == int)

            subscription = yield handler.subscribe(
                on_event,
                "com.myapp.topic1",
                options=types.SubscribeOptions(match="wildcard", get_retained=True),
            )
            self.assertTrue(type(subscription.id) == int)

        @inlineCallbacks
        def test_double_subscribe(self):
            handler = ApplicationSession()
            MockTransport(handler)

            event0 = Deferred()
            event1 = Deferred()

            subscription0 = yield handler.subscribe(
                lambda: event0.callback(42), "com.myapp.topic1"
            )
            subscription1 = yield handler.subscribe(
                lambda: event1.callback("foo"), "com.myapp.topic1"
            )
            # same topic, same ID
            self.assertTrue(subscription0.id == subscription1.id)

            # do a publish (MockTransport fakes the acknowledgement
            # message) and then do an actual publish event. The IDs
            # are the same, so we just do one Event.
            publish = yield handler.publish(
                "com.myapp.topic1",
                options=types.PublishOptions(acknowledge=True, exclude_me=False),
            )
            handler.onMessage(message.Event(subscription0.id, publish.id))

            # ensure we actually got both callbacks
            self.assertTrue(event0.called, "Missing callback")
            self.assertTrue(event1.called, "Missing callback")

        @inlineCallbacks
        def test_double_subscribe_single_unsubscribe(self):
            """
            Make sure we correctly deal with unsubscribing one of our handlers
            from the same topic.
            """
            handler = ApplicationSession()
            MockTransport(handler)

            # monkey-patch ApplicationSession to ensure we DO NOT get
            # an Unsubscribed message -- since we only unsubscribe
            # from ONE of our handlers for com.myapp.topic1
            def onMessage(msg):
                assert not isinstance(msg, message.Unsubscribed)
                return ApplicationSession.onMessage(handler, msg)

            handler.onMessage = onMessage

            event0 = Deferred()
            event1 = Deferred()

            subscription0 = yield handler.subscribe(
                lambda: event0.callback(42), "com.myapp.topic1"
            )
            subscription1 = yield handler.subscribe(
                lambda: event1.callback("foo"), "com.myapp.topic1"
            )

            self.assertTrue(subscription0.id == subscription1.id)
            yield subscription1.unsubscribe()

            # do a publish (MockTransport fakes the acknowledgement
            # message) and then do an actual publish event. Note the
            # IDs are the same, so there's only one Event.
            publish = yield handler.publish(
                "com.myapp.topic1",
                options=types.PublishOptions(acknowledge=True, exclude_me=False),
            )
            handler.onMessage(message.Event(subscription0.id, publish.id))

            # since we unsubscribed the second event handler, we
            # should NOT have called its callback
            self.assertTrue(event0.called, "Missing callback")
            self.assertTrue(not event1.called, "Second callback fired.")

        @inlineCallbacks
        def test_double_subscribe_double_unsubscribe(self):
            """
            If we subscribe twice, and unsubscribe twice, we should then get
            an Unsubscribed message.
            """
            handler = ApplicationSession()
            MockTransport(handler)

            # monkey-patch ApplicationSession to ensure we get our message
            unsubscribed_d = Deferred()

            def onMessage(msg):
                if isinstance(msg, message.Unsubscribed):
                    unsubscribed_d.callback(msg)
                return ApplicationSession.onMessage(handler, msg)

            handler.onMessage = onMessage

            event0 = Deferred()
            event1 = Deferred()

            subscription0 = yield handler.subscribe(
                lambda: event0.callback(42), "com.myapp.topic1"
            )
            subscription1 = yield handler.subscribe(
                lambda: event1.callback("foo"), "com.myapp.topic1"
            )

            self.assertTrue(subscription0.id == subscription1.id)
            yield subscription0.unsubscribe()
            yield subscription1.unsubscribe()

            # after the second unsubscribe, we should have gotten an
            # Unsubscribed message
            assert unsubscribed_d.called

            # do a publish (MockTransport fakes the acknowledgement
            # message) and then do an actual publish event. Sending
            # the Event should be an error, as we have no
            # subscriptions left.
            publish = yield handler.publish(
                "com.myapp.topic1",
                options=types.PublishOptions(acknowledge=True, exclude_me=False),
            )
            try:
                handler.onMessage(message.Event(subscription0.id, publish.id))
                self.fail("Expected ProtocolError")
            except ProtocolError:
                pass

            # since we unsubscribed the second event handler, we
            # should NOT have called its callback
            self.assertTrue(not event0.called, "First callback fired.")
            self.assertTrue(not event1.called, "Second callback fired.")

        @inlineCallbacks
        def test_double_subscribe_errors(self):
            """
            Test various error-conditions when we try to add a second
            subscription-handler (its signature must match any
            existing handlers).
            """
            handler = ApplicationSession()
            MockTransport(handler)

            event0 = Deferred()
            event1 = Deferred()

            def second(*args, **kw):
                # our EventDetails should have been passed as the
                # "boom" kwarg; see "details_arg=" below
                self.assertTrue("boom" in kw)
                self.assertTrue(isinstance(kw["boom"], types.EventDetails))
                event1.callback(args)

            subscription0 = yield handler.subscribe(
                lambda arg: event0.callback(arg), "com.myapp.topic1"
            )
            subscription1 = yield handler.subscribe(
                second,
                "com.myapp.topic1",
                types.SubscribeOptions(details_arg="boom"),
            )
            # same topic, same ID
            self.assertTrue(subscription0.id == subscription1.id)

            # MockTransport gives us the ack reply and then we do our
            # own event message.
            publish = yield handler.publish(
                "com.myapp.topic1",
                options=types.PublishOptions(acknowledge=True, exclude_me=False),
            )
            # note that the protocol serializer converts all sequences
            # to lists, so we pass "args" as a list, not a tuple on
            # purpose.
            handler.onMessage(
                message.Event(subscription0.id, publish.id, args=["arg0"])
            )

            # each callback should have gotten called, each with its
            # own args (we check the correct kwarg in second() above)
            self.assertTrue(event0.called)
            self.assertTrue(event1.called)
            self.assertEqual(event0.result, "arg0")
            self.assertEqual(event1.result, ("arg0",))

        @inlineCallbacks
        def test_publish_callback_exception(self):
            """
            Ensure we handle an exception from the user code.
            """
            handler = ApplicationSession()
            MockTransport(handler)

            error_instance = RuntimeError("we have a problem")
            got_err_d = Deferred()

            def observer(e, msg):
                if error_instance == e.value:
                    got_err_d.callback(True)

            handler.onUserError = observer

            def boom():
                raise error_instance

            sub = yield handler.subscribe(boom, "com.myapp.topic1")

            # MockTransport gives us the ack reply and then we do our
            # own event message
            publish = yield handler.publish(
                "com.myapp.topic1",
                options=types.PublishOptions(acknowledge=True, exclude_me=False),
            )
            msg = message.Event(sub.id, publish.id)
            handler.onMessage(msg)

            # we know it worked if our observer worked and did
            # .callback on our Deferred above.
            self.assertTrue(got_err_d.called)

        @inlineCallbacks
        def test_unsubscribe(self):
            handler = ApplicationSession()
            MockTransport(handler)

            def on_event(*args, **kwargs):
                print("got event", args, kwargs)

            subscription = yield handler.subscribe(on_event, "com.myapp.topic1")
            yield subscription.unsubscribe()

        @inlineCallbacks
        def test_register(self):
            handler = ApplicationSession()
            MockTransport(handler)

            def on_call(*args, **kwargs):
                print("got call", args, kwargs)

            registration = yield handler.register(on_call, "com.myapp.procedure1")
            self.assertTrue(type(registration.id) == int)

            registration = yield handler.register(
                on_call,
                "com.myapp.procedure1",
                options=types.RegisterOptions(match="prefix"),
            )
            self.assertTrue(type(registration.id) == int)

        @inlineCallbacks
        def test_unregister(self):
            handler = ApplicationSession()
            MockTransport(handler)

            def on_call(*args, **kwargs):
                print("got call", args, kwargs)

            registration = yield handler.register(on_call, "com.myapp.procedure1")
            yield registration.unregister()

        def test_unregister_no_such_registration(self):
            if twisted.version.major < 14:
                return
            handler = ApplicationSession()
            transport = MockTransport(handler)

            with self.assertRaises(ProtocolError) as ctx:
                transport.send(message.Unregister(0, 1234))
            self.assertIn(
                "UNREGISTERED received for non-existant registration",
                str(ctx.exception),
            )

        @inlineCallbacks
        def test_unregister_log(self):
            handler = ApplicationSession()
            transport = MockTransport(handler)

            def on_call(*args, **kwargs):
                on_call.called = True

            on_call.called = False

            registration = yield handler.register(on_call, "com.myapp.procedure1")
            transport.drop_registration(registration.id)
            self.assertFalse(on_call.called)

        def test_on_disconnect_error(self):
            errors = []

            class AppSess(ApplicationSession):
                def onUserError(self, e, msg):
                    errors.append((e.value, msg))

                def onDisconnect(self, foo, bar, quux="snark"):
                    # this over-ridden onDisconnect takes the wrong args
                    raise RuntimeError("This shouldn't happen")

            s = AppSess()
            MockTransport(s)

            # when the transport closes, it calls onClose which should
            # dispatch onDisconnect
            s.onClose(False)

            # ...but we should collect an error, because our
            # onDisconnect takes the wrong arguments.
            self.assertEqual(len(errors), 1)
            self.assertTrue(isinstance(errors[0][0], TypeError))

    class TestInvoker(unittest.TestCase):
        @inlineCallbacks
        def test_invoke(self):
            handler = ApplicationSession()
            MockTransport(handler)

            def myproc1():
                return 23

            yield handler.register(myproc1, "com.myapp.myproc1")

            res = yield handler.call("com.myapp.myproc1")
            self.assertEqual(res, 23)

        @inlineCallbacks
        def test_invoke_twice(self):
            handler = ApplicationSession()
            MockTransport(handler)

            def myproc1():
                return 23

            yield handler.register(myproc1, "com.myapp.myproc1")

            d0 = handler.call("com.myapp.myproc1")
            d1 = handler.call("com.myapp.myproc1")
            res = yield DeferredList([d0, d1])
            self.assertEqual(res, [(True, 23), (True, 23)])

        @inlineCallbacks
        def test_invoke_request_id_sequences(self):
            """
            make sure each session independently generates sequential IDs
            """
            handler0 = ApplicationSession()
            handler1 = ApplicationSession()
            trans0 = MockTransport(handler0)
            trans1 = MockTransport(handler1)

            # the ID sequences for each session should both start at 0
            # (the register) and then increment for the call()
            def verify_seq_id(orig, msg):
                if isinstance(msg, message.Register):
                    self.assertEqual(msg.request, 1)
                elif isinstance(msg, message.Call):
                    self.assertEqual(msg.request, 2)
                return orig(msg)

            orig0 = trans0.send
            orig1 = trans1.send
            trans0.send = lambda msg: verify_seq_id(orig0, msg)
            trans1.send = lambda msg: verify_seq_id(orig1, msg)

            def myproc1():
                return 23

            yield handler0.register(myproc1, "com.myapp.myproc1")
            yield handler1.register(myproc1, "com.myapp.myproc1")

            d0 = handler0.call("com.myapp.myproc1")
            d1 = handler1.call("com.myapp.myproc1")
            res = yield DeferredList([d0, d1])
            self.assertEqual(res, [(True, 23), (True, 23)])

        @inlineCallbacks
        def test_invoke_user_raises(self):
            handler = ApplicationSession()
            handler.traceback_app = True
            MockTransport(handler)
            errors = []

            def log_error(e, msg):
                errors.append((e.value, msg))

            handler.onUserError = log_error

            name_error = NameError("foo")

            def bing():
                raise name_error

            # see MockTransport, must start with "com.myapp.myproc"
            yield handler.register(bing, "com.myapp.myproc99")

            try:
                yield handler.call("com.myapp.myproc99")
                self.fail("Expected an error")
            except Exception as e:
                # XXX should/could we export all the builtin types?
                # right now, we always get ApplicationError
                # self.assertTrue(isinstance(e, NameError))
                self.assertTrue(isinstance(e, RuntimeError))

            # also, we should have logged the real NameError to
            # Twisted.
            self.assertEqual(1, len(errors))
            self.assertEqual(name_error, errors[0][0])

        @inlineCallbacks
        def test_invoke_progressive_result(self):
            handler = ApplicationSession()
            MockTransport(handler)

            @inlineCallbacks
            def bing(details=None):
                self.assertTrue(details is not None)
                self.assertTrue(details.progress is not None)
                for i in range(10):
                    details.progress(i)
                    yield succeed(i)
                return 42

            progressive = list(map(lambda _: Deferred(), range(10)))

            def progress(arg):
                progressive[arg].callback(arg)

            # see MockTransport, must start with "com.myapp.myproc"
            yield handler.register(
                bing,
                "com.myapp.myproc2",
                types.RegisterOptions(details_arg="details"),
            )
            res = yield handler.call(
                "com.myapp.myproc2",
                options=types.CallOptions(on_progress=progress),
            )
            self.assertEqual(42, res)
            # make sure we got *all* our progressive results
            for i in range(10):
                self.assertTrue(progressive[i].called)
                self.assertEqual(i, progressive[i].result)

        @inlineCallbacks
        def test_invoke_progressive_result_error(self):
            handler = ApplicationSession()
            MockTransport(handler)

            @inlineCallbacks
            def bing(arg, details=None, key=None):
                self.assertTrue(details is not None)
                self.assertTrue(details.progress is not None)
                self.assertEqual(key, "word")
                self.assertEqual("arg", arg)
                details.progress("life", something="nothing")
                yield succeed("meaning of")
                return 42

            got_progress = Deferred()
            progress_error = NameError("foo")
            logged_errors = []

            def got_error(e, msg):
                logged_errors.append((e.value, msg))

            handler.onUserError = got_error

            def progress(arg, something=None):
                self.assertEqual("nothing", something)
                got_progress.callback(arg)
                raise progress_error

            # see MockTransport, must start with "com.myapp.myproc"
            yield handler.register(
                bing,
                "com.myapp.myproc2",
                types.RegisterOptions(details_arg="details"),
            )

            res = yield handler.call(
                "com.myapp.myproc2",
                "arg",
                options=types.CallOptions(on_progress=progress),
                key="word",
            )

            self.assertEqual(42, res)
            # our progress handler raised an error, but not before
            # recording success.
            self.assertTrue(got_progress.called)
            self.assertEqual("life", got_progress.result)
            # make sure our progress-handler error was logged
            self.assertEqual(1, len(logged_errors))
            self.assertEqual(progress_error, logged_errors[0][0])

        @inlineCallbacks
        def test_invoke_progressive_result_no_args(self):
            handler = ApplicationSession()
            MockTransport(handler)

            @inlineCallbacks
            def bing(details=None):
                self.assertTrue(details is not None)
                self.assertTrue(details.progress is not None)
                details.progress()
                yield succeed(True)
                return 42

            got_progress = Deferred()

            def progress():
                got_progress.callback("intentionally left blank")

            # see MockTransport, must start with "com.myapp.myproc"
            yield handler.register(
                bing,
                "com.myapp.myproc2",
                types.RegisterOptions(details_arg="details"),
            )

            res = yield handler.call(
                "com.myapp.myproc2",
                options=types.CallOptions(on_progress=progress),
            )
            self.assertEqual(42, res)
            self.assertTrue(got_progress.called)

        @inlineCallbacks
        def test_invoke_progressive_result_just_kwargs(self):
            handler = ApplicationSession()
            MockTransport(handler)

            @inlineCallbacks
            def bing(details=None):
                self.assertTrue(details is not None)
                self.assertTrue(details.progress is not None)
                details.progress(key="word")
                yield succeed(True)
                return 42

            got_progress = Deferred()

            def progress(key=None):
                got_progress.callback(key)

            # see MockTransport, must start with "com.myapp.myproc"
            yield handler.register(
                bing,
                "com.myapp.myproc2",
                types.RegisterOptions(details_arg="details"),
            )

            res = yield handler.call(
                "com.myapp.myproc2",
                options=types.CallOptions(on_progress=progress),
            )
            self.assertEqual(42, res)
            self.assertTrue(got_progress.called)
            self.assertEqual("word", got_progress.result)

        @inlineCallbacks
        def test_call_exception_runtimeerror(self):
            handler = ApplicationSession()
            MockTransport(handler)
            exception = RuntimeError("a simple error")

            def raiser():
                raise exception

            registration0 = yield handler.register(raiser, "com.myapp.myproc_error")
            try:
                yield handler.call("com.myapp.myproc_error")
                self.fail()
            except Exception as e:
                self.assertIsInstance(e, ApplicationError)
                self.assertEqual(
                    e.error_message(), "wamp.error.runtime_error: a simple error"
                )
            finally:
                yield registration0.unregister()

        @inlineCallbacks
        def test_call_exception_bare(self):
            handler = ApplicationSession()
            MockTransport(handler)
            exception = Exception()

            def raiser():
                raise exception

            registration0 = yield handler.register(raiser, "com.myapp.myproc_error")
            try:
                yield handler.call("com.myapp.myproc_error")
                self.fail()
            except Exception as e:
                self.assertIsInstance(e, ApplicationError)
            finally:
                yield registration0.unregister()

        # ## variant 1: works
        # def test_publish1(self):
        #    d = self.handler.publish('de.myapp.topic1')
        #    self.assertFailure(d, ApplicationError)

        # ## variant 2: works
        # @inlineCallbacks
        # def test_publish2(self):
        #    yield self.assertFailure(self.handler.publish('de.myapp.topic1'), ApplicationError)

        # ## variant 3: does NOT work
        # @inlineCallbacks
        # def test_publish3(self):
        #    with self.assertRaises(ApplicationError):
        #       yield self.handler.publish('de.myapp.topic1')

    class TestAuthenticator(unittest.TestCase):
        def test_inconsistent_authids(self):
            session = Session(mock.Mock())
            auth0 = create_authenticator(
                "wampcra",
                authid="alice",
                secret="p4ssw0rd",
            )
            auth1 = create_authenticator(
                "wampcra",
                authid="bob",
                secret="password42",
            )

            session.add_authenticator(auth0)
            with self.assertRaises(ValueError) as ctx:
                session.add_authenticator(auth1)
            assert "authids" in str(ctx.exception)

        def test_two_authenticators(self):
            session = Session(mock.Mock())

            class TestAuthenticator(IAuthenticator):
                name = "test"

                def on_challenge(self, session, challenge):
                    raise NotImplementedError

                def on_welcome(self, authextra):
                    raise NotImplementedError

            auth0 = TestAuthenticator()
            auth0.authextra = {
                "foo": "value0",
                "bar": "value1",
            }
            auth0._args = {}

            auth1 = TestAuthenticator()
            auth1.authextra = {
                "bar": "value1",
                "qux": "what",
            }
            auth1._args = {}

            session.add_authenticator(auth0)
            session.add_authenticator(auth1)

        def test_inconsistent_authextra(self):
            session = Session(mock.Mock())

            class TestAuthenticator(IAuthenticator):
                name = "test"

                def on_challenge(self, session, challenge):
                    raise NotImplementedError

                def on_welcome(self, authextra):
                    raise NotImplementedError

            auth0 = TestAuthenticator()
            auth0.authextra = {
                "foo": "value0",
                "bar": "value1",
            }
            auth0._args = {}

            auth1 = TestAuthenticator()
            auth1.authextra = {
                "foo": "value1",
            }
            auth1._args = {}

            session.add_authenticator(auth0)
            with self.assertRaises(ValueError) as ctx:
                session.add_authenticator(auth1)
            self.assertIn("Inconsistent authextra", str(ctx.exception))



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


# WAMP session identity information.
class Principal(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Principal()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsPrincipal(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Principal
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # WAMP session ID.
    # Principal
    def Session(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # WAMP session authentication ID.
    # Principal
    def Authid(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # WAMP session authentication role.
    # Principal
    def Authrole(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None


def PrincipalStart(builder):
    builder.StartObject(3)


def Start(builder):
    PrincipalStart(builder)


def PrincipalAddSession(builder, session):
    builder.PrependUint64Slot(0, session, 0)


def AddSession(builder, session):
    PrincipalAddSession(builder, session)


def PrincipalAddAuthid(builder, authid):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(authid), 0
    )


def AddAuthid(builder, authid):
    PrincipalAddAuthid(builder, authid)


def PrincipalAddAuthrole(builder, authrole):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(authrole), 0
    )


def AddAuthrole(builder, authrole):
    PrincipalAddAuthrole(builder, authrole)


def PrincipalEnd(builder):
    return builder.EndObject()


def End(builder):
    return PrincipalEnd(builder)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import unittest

# from .test_wamp_transport_details import TRANSPORT_DETAILS_1
from autobahn.wamp.test.test_wamp_transport_details import TRANSPORT_DETAILS_1
from autobahn.wamp.types import SessionDetails, TransportDetails


class TestSessionDetails(unittest.TestCase):
    def test_empty(self):
        sd1 = SessionDetails()
        data = sd1.marshal()
        self.assertEqual(
            data,
            {
                "realm": None,
                "session": None,
                "authid": None,
                "authrole": None,
                "authmethod": None,
                "authprovider": None,
                "authextra": None,
                "serializer": None,
                "transport": None,
                "resumed": None,
                "resumable": None,
                "resume_token": None,
            },
        )
        sd2 = SessionDetails.parse(data)
        self.assertEqual(sd2, sd1)

    def test_attributes(self):
        sd1 = SessionDetails()
        td1 = TransportDetails.parse(TRANSPORT_DETAILS_1)
        sd1.realm = "realm1"
        sd1.session = 666
        sd1.authid = "homer"
        sd1.authrole = "user"
        sd1.authmethod = "wampcra"
        sd1.authprovider = "static"
        sd1.authextra = {"foo": "bar", "baz": [1, 2, 3]}
        sd1.serializer = "json"
        sd1.transport = td1
        sd1.resumed = False
        sd1.resumable = True
        sd1.resume_token = "8713e25a-d4f5-48b7-9d6d-eda66603a1ab"
        data = sd1.marshal()
        self.assertEqual(
            data,
            {
                "realm": sd1.realm,
                "session": sd1.session,
                "authid": sd1.authid,
                "authrole": sd1.authrole,
                "authmethod": sd1.authmethod,
                "authprovider": sd1.authprovider,
                "authextra": sd1.authextra,
                "serializer": sd1.serializer,
                "transport": sd1.transport.marshal(),
                "resumed": sd1.resumed,
                "resumable": sd1.resumable,
                "resume_token": sd1.resume_token,
            },
        )
        sd2 = SessionDetails.parse(data)
        self.assertEqual(sd2, sd1)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class PublisherFeatures(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = PublisherFeatures()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsPublisherFeatures(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # PublisherFeatures
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # PublisherFeatures
    def PublisherIdentification(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # PublisherFeatures
    def PublisherExclusion(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # PublisherFeatures
    def SubscriberBlackwhiteListing(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # PublisherFeatures
    def AcknowledgeEventReceived(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # PublisherFeatures
    def PayloadTransparency(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # PublisherFeatures
    def PayloadEncryptionCryptobox(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False


def PublisherFeaturesStart(builder):
    builder.StartObject(6)


def Start(builder):
    PublisherFeaturesStart(builder)


def PublisherFeaturesAddPublisherIdentification(builder, publisherIdentification):
    builder.PrependBoolSlot(0, publisherIdentification, 0)


def AddPublisherIdentification(builder, publisherIdentification):
    PublisherFeaturesAddPublisherIdentification(builder, publisherIdentification)


def PublisherFeaturesAddPublisherExclusion(builder, publisherExclusion):
    builder.PrependBoolSlot(1, publisherExclusion, 0)


def AddPublisherExclusion(builder, publisherExclusion):
    PublisherFeaturesAddPublisherExclusion(builder, publisherExclusion)


def PublisherFeaturesAddSubscriberBlackwhiteListing(
    builder, subscriberBlackwhiteListing
):
    builder.PrependBoolSlot(2, subscriberBlackwhiteListing, 0)


def AddSubscriberBlackwhiteListing(builder, subscriberBlackwhiteListing):
    PublisherFeaturesAddSubscriberBlackwhiteListing(
        builder, subscriberBlackwhiteListing
    )


def PublisherFeaturesAddAcknowledgeEventReceived(builder, acknowledgeEventReceived):
    builder.PrependBoolSlot(3, acknowledgeEventReceived, 0)


def AddAcknowledgeEventReceived(builder, acknowledgeEventReceived):
    PublisherFeaturesAddAcknowledgeEventReceived(builder, acknowledgeEventReceived)


def PublisherFeaturesAddPayloadTransparency(builder, payloadTransparency):
    builder.PrependBoolSlot(4, payloadTransparency, 0)


def AddPayloadTransparency(builder, payloadTransparency):
    PublisherFeaturesAddPayloadTransparency(builder, payloadTransparency)


def PublisherFeaturesAddPayloadEncryptionCryptobox(builder, payloadEncryptionCryptobox):
    builder.PrependBoolSlot(5, payloadEncryptionCryptobox, 0)


def AddPayloadEncryptionCryptobox(builder, payloadEncryptionCryptobox):
    PublisherFeaturesAddPayloadEncryptionCryptobox(builder, payloadEncryptionCryptobox)


def PublisherFeaturesEnd(builder):
    return builder.EndObject()


def End(builder):
    return PublisherFeaturesEnd(builder)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import platform

import autobahn
from autobahn.websocket import USES_NVX

# WAMP support
from autobahn.asyncio.wamp import ApplicationSession

# WebSocket protocol support
from autobahn.asyncio.websocket import (
    WebSocketClientFactory,
    WebSocketClientProtocol,
    WebSocketServerFactory,
    WebSocketServerProtocol,
)

__all__ = (
    "ApplicationSession",
    "WebSocketClientFactory",
    "WebSocketClientProtocol",
    "WebSocketServerFactory",
    "WebSocketServerProtocol",
)

# Build identification string with optional NVX acceleration indicator
if USES_NVX:
    import cffi

    __ident__ = "Autobahn/{}-NVXCFFI/{}-asyncio-{}/{}".format(
        autobahn.__version__,
        cffi.__version__,
        platform.python_implementation(),
        platform.python_version(),
    )
else:
    __ident__ = "Autobahn/{}-asyncio-{}/{}".format(
        autobahn.__version__,
        platform.python_implementation(),
        platform.python_version(),
    )

"""
Identification string for the Autobahn|Python asyncio backend.

This string identifies the library version, networking framework, and runtime
environment. It's commonly used in protocol handshakes (e.g., WebSocket Upgrade
headers, WAMP HELLO metadata) to identify the client/server implementation.

Format with NVX acceleration enabled::

    "Autobahn/{version}-NVXCFFI/{cffi_version}-asyncio-{python_impl}/{python_version}"

Format without NVX acceleration::

    "Autobahn/{version}-asyncio-{python_impl}/{python_version}"

The presence of ``NVXCFFI`` in the identification string indicates that NVX
(Native Vector Extensions) native acceleration is enabled, providing high-performance
UTF-8 validation and XOR masking using SIMD instructions.

Note that asyncio is built into Python's standard library (since Python 3.4),
so no separate asyncio version is included in the identification string (unlike
the Twisted backend which includes the Twisted version).

:type: str

Example values::

    # With NVX acceleration on CPython
    "Autobahn/25.10.2-NVXCFFI/1.15.1-asyncio-CPython/3.11.9"

    # Without NVX acceleration on PyPy
    "Autobahn/25.10.2-asyncio-PyPy/7.3.16"

See Also:
    :data:`autobahn.twisted.__ident__` - Identification string for Twisted backend
    :data:`autobahn.websocket.USES_NVX` - Whether NVX acceleration is enabled
"""



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class Published(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Published()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsPublished(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # Published
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Published
    def Session(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # Published
    def Request(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # Published
    def Publication(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0


def PublishedStart(builder):
    builder.StartObject(3)


def Start(builder):
    PublishedStart(builder)


def PublishedAddSession(builder, session):
    builder.PrependUint64Slot(0, session, 0)


def AddSession(builder, session):
    PublishedAddSession(builder, session)


def PublishedAddRequest(builder, request):
    builder.PrependUint64Slot(1, request, 0)


def AddRequest(builder, request):
    PublishedAddRequest(builder, request)


def PublishedAddPublication(builder, publication):
    builder.PrependUint64Slot(2, publication, 0)


def AddPublication(builder, publication):
    PublishedAddPublication(builder, publication)


def PublishedEnd(builder):
    return builder.EndObject()


def End(builder):
    return PublishedEnd(builder)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import os
import unittest
from base64 import b64encode
from hashlib import sha1
from unittest.mock import Mock

import txaio

from autobahn.testutil import FakeTransport
from autobahn.wamp.types import TransportDetails
from autobahn.websocket.protocol import (
    WebSocketClientFactory,
    WebSocketClientProtocol,
    WebSocketProtocol,
    WebSocketServerFactory,
    WebSocketServerProtocol,
)
from autobahn.websocket.types import ConnectingRequest


class WebSocketClientProtocolTests(unittest.TestCase):
    def setUp(self):
        t = FakeTransport()
        f = WebSocketClientFactory()
        f.log = txaio.make_logger()
        p = WebSocketClientProtocol()
        p.log = txaio.make_logger()
        p.factory = f
        p.transport = t
        p._transport_details = TransportDetails()

        p._connectionMade()
        p.state = p.STATE_OPEN
        p.websocket_version = 18

        self.protocol = p
        self.transport = t

    def tearDown(self):
        for call in [
            self.protocol.autoPingPendingCall,
            self.protocol.autoPingTimeoutCall,
            self.protocol.openHandshakeTimeoutCall,
            self.protocol.closeHandshakeTimeoutCall,
        ]:
            if call is not None:
                call.cancel()

    def test_auto_ping(self):
        self.protocol.autoPingInterval = 1
        self.protocol.websocket_protocols = [Mock()]
        self.protocol.websocket_extensions = []
        self.protocol._onOpen = lambda: None
        self.protocol._wskey = "0" * 24
        self.protocol.peer = Mock()

        # usually provided by the Twisted or asyncio specific
        # subclass, but we're testing the parent here...
        self.protocol._onConnect = Mock()
        self.protocol._closeConnection = Mock()

        # set up a connection
        self.protocol._actuallyStartHandshake(
            ConnectingRequest(
                host="example.com",
                port=80,
                resource="/ws",
            )
        )

        key = self.protocol.websocket_key + WebSocketProtocol._WS_MAGIC
        self.protocol.data = (
            b"HTTP/1.1 101 Switching Protocols\x0d\x0a"
            b"Upgrade: websocket\x0d\x0a"
            b"Connection: upgrade\x0d\x0a"
            b"Sec-Websocket-Accept: "
            + b64encode(sha1(key).digest())
            + b"\x0d\x0a\x0d\x0a"
        )
        self.protocol.processHandshake()

        self.assertTrue(self.protocol.autoPingPendingCall is not None)


class WebSocketServerProtocolTests(unittest.TestCase):
    """
    Tests for autobahn.websocket.protocol.WebSocketProtocol.
    """

    def setUp(self):
        t = FakeTransport()
        f = WebSocketServerFactory()
        f.log = txaio.make_logger()
        p = WebSocketServerProtocol()
        p.log = txaio.make_logger()
        p.factory = f
        p.transport = t

        p._connectionMade()
        p.state = p.STATE_OPEN
        p.websocket_version = 18

        self.protocol = p
        self.transport = t

    def tearDown(self):
        for call in [
            self.protocol.autoPingPendingCall,
            self.protocol.autoPingTimeoutCall,
            self.protocol.openHandshakeTimeoutCall,
            self.protocol.closeHandshakeTimeoutCall,
        ]:
            if call is not None:
                call.cancel()

    def test_auto_ping(self):
        proto = Mock()
        proto._get_seconds = Mock(return_value=1)
        self.protocol.autoPingInterval = 1
        self.protocol.websocket_protocols = [proto]
        self.protocol.websocket_extensions = []
        self.protocol._onOpen = lambda: None
        self.protocol._wskey = "0" * 24
        self.protocol.succeedHandshake(proto)

        self.assertTrue(self.protocol.autoPingPendingCall is not None)

    def test_sendClose_none(self):
        """
        sendClose with no code or reason works.
        """
        self.protocol.sendClose()

        # We closed properly
        self.assertEqual(self.transport._written, b"\x88\x00")
        self.assertEqual(self.protocol.state, self.protocol.STATE_CLOSING)

    def test_sendClose_str_reason(self):
        """
        sendClose with a str reason works.
        """
        self.protocol.sendClose(code=1000, reason="oh no")

        # We closed properly
        self.assertEqual(self.transport._written[2:], b"\x03\xe8oh no")
        self.assertEqual(self.protocol.state, self.protocol.STATE_CLOSING)

    def test_sendClose_unicode_reason(self):
        """
        sendClose with a unicode reason works.
        """
        self.protocol.sendClose(code=1000, reason="oh no")

        # We closed properly
        self.assertEqual(self.transport._written[2:], b"\x03\xe8oh no")
        self.assertEqual(self.protocol.state, self.protocol.STATE_CLOSING)

    def test_sendClose_toolong(self):
        """
        sendClose with a too-long reason will truncate it.
        """
        self.protocol.sendClose(code=1000, reason="abc" * 1000)

        # We closed properly
        self.assertEqual(self.transport._written[2:], b"\x03\xe8" + (b"abc" * 41))
        self.assertEqual(self.protocol.state, self.protocol.STATE_CLOSING)

    def test_sendClose_reason_with_no_code(self):
        """
        Trying to sendClose with a reason but no code will raise an Exception.
        """
        with self.assertRaises(Exception) as e:
            self.protocol.sendClose(reason="abc")

        self.assertIn("close reason without close code", str(e.exception))

        # We shouldn't have closed
        self.assertEqual(self.transport._written, b"")
        self.assertEqual(self.protocol.state, self.protocol.STATE_OPEN)

    def test_sendClose_invalid_code_type(self):
        """
        Trying to sendClose with a non-int code will raise an Exception.
        """
        with self.assertRaises(Exception) as e:
            self.protocol.sendClose(code="134")

        self.assertIn("invalid type", str(e.exception))

        # We shouldn't have closed
        self.assertEqual(self.transport._written, b"")
        self.assertEqual(self.protocol.state, self.protocol.STATE_OPEN)

    def test_sendClose_invalid_code_value(self):
        """
        Trying to sendClose with a non-valid int code will raise an Exception.
        """
        with self.assertRaises(Exception) as e:
            self.protocol.sendClose(code=10)

        self.assertIn("invalid close code 10", str(e.exception))

        # We shouldn't have closed
        self.assertEqual(self.transport._written, b"")
        self.assertEqual(self.protocol.state, self.protocol.STATE_OPEN)

    def test_interpolate_server_status_template(self):
        from autobahn.websocket.protocol import _SERVER_STATUS_TEMPLATE

        s = _SERVER_STATUS_TEMPLATE.format(None, "0.0.0")
        self.assertEqual(type(s), str)
        self.assertTrue(len(s) > 0)


if os.environ.get("USE_TWISTED", False):

    class TwistedProtocolTests(unittest.TestCase):
        """
        Tests which require a specific framework's protocol class to work
        (in this case, using Twisted)
        """

        def setUp(self):
            from autobahn.twisted.websocket import (
                WebSocketServerFactory,
                WebSocketServerProtocol,
            )

            t = FakeTransport()
            f = WebSocketServerFactory()
            p = WebSocketServerProtocol()
            p.factory = f
            p.transport = t

            p._connectionMade()
            p.state = p.STATE_OPEN
            p.websocket_version = 18

            self.protocol = p
            self.transport = t

        def tearDown(self):
            for call in [
                self.protocol.autoPingPendingCall,
                self.protocol.autoPingTimeoutCall,
                self.protocol.openHandshakeTimeoutCall,
                self.protocol.closeHandshakeTimeoutCall,
            ]:
                if call is not None:
                    call.cancel()

        def test_loseConnection(self):
            """
            If we lose our connection before openHandshakeTimeout fires, it is
            cleaned up
            """
            # so, I guess a little cheezy, but we depend on the asyncio or
            # twisted class to call _connectionLost at some point; faking
            # that here
            self.protocol._connectionLost(txaio.create_failure(RuntimeError("testing")))
            self.assertTrue(self.protocol.openHandshakeTimeoutCall is None)

        def test_send_server_status(self):
            self.protocol.sendServerStatus()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# WAMP transport channel type (`channel_type`). WAMP can run over any Transport which is message-based (requiring no WAMP transport channel framing) or stream-based (requiring WAMP transport channel framing), and bidirectional, reliable and ordered.
class TransportChannelType(object):
    # Not set or applicable.
    NULL = 0
    # Host language native function call transport, e.g. inherently message-based already (function call).
    FUNCTION = 1
    # Host run-time / OS level in-memory transport, e.g. (unframed / stream-based) memory buffer.
    MEMORY = 2
    # Serial (UART) based transport (unframed / stream-based).
    SERIAL = 3
    # TCP (non-TLS) based transport (unframed / stream-based).
    TCP = 4
    # TLS (over TCP) based transport (unframed / stream-based).
    TLS = 5



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class AuthScramChallenge(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = AuthScramChallenge()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsAuthScramChallenge(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # AuthScramChallenge
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # AuthScramChallenge
    def Nonce(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # AuthScramChallenge
    def Salt(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # AuthScramChallenge
    def Kdf(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 1

    # AuthScramChallenge
    def Iterations(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags, o + self._tab.Pos
            )
        return 0

    # AuthScramChallenge
    def Memory(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags, o + self._tab.Pos
            )
        return 0

    # AuthScramChallenge
    def ChannelBinding(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 0


def AuthScramChallengeStart(builder):
    builder.StartObject(6)


def Start(builder):
    AuthScramChallengeStart(builder)


def AuthScramChallengeAddNonce(builder, nonce):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(nonce), 0
    )


def AddNonce(builder, nonce):
    AuthScramChallengeAddNonce(builder, nonce)


def AuthScramChallengeAddSalt(builder, salt):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(salt), 0
    )


def AddSalt(builder, salt):
    AuthScramChallengeAddSalt(builder, salt)


def AuthScramChallengeAddKdf(builder, kdf):
    builder.PrependUint8Slot(2, kdf, 1)


def AddKdf(builder, kdf):
    AuthScramChallengeAddKdf(builder, kdf)


def AuthScramChallengeAddIterations(builder, iterations):
    builder.PrependUint32Slot(3, iterations, 0)


def AddIterations(builder, iterations):
    AuthScramChallengeAddIterations(builder, iterations)


def AuthScramChallengeAddMemory(builder, memory):
    builder.PrependUint32Slot(4, memory, 0)


def AddMemory(builder, memory):
    AuthScramChallengeAddMemory(builder, memory)


def AuthScramChallengeAddChannelBinding(builder, channelBinding):
    builder.PrependUint8Slot(5, channelBinding, 0)


def AddChannelBinding(builder, channelBinding):
    AuthScramChallengeAddChannelBinding(builder, channelBinding)


def AuthScramChallengeEnd(builder):
    return builder.EndObject()


def End(builder):
    return AuthScramChallengeEnd(builder)
